"""Seismic Unix trace header mappings.

We added aliases for SEG-Y Rev0 equivalent fields.

Aliases for extended SU headers (after Rev0 before Rev1) are not
included because the types of them don't match SEG-Y Rev1, and we
don't want to point to corrupt data if user expects SU type data.

Source:
https://github-wiki-see.page/m/JohnWStockwellJr/SeisUnix/wiki/Seismic-Unix-data-format
"""

from segy.standards.fields import trace

# fmt: off
SEIS_UNIX_TRACE_FIELD_MAP = {
    "tracl": trace.Rev0.TRACE_SEQ_NUM_LINE,
    "tracr": trace.Rev0.TRACE_SEQ_NUM_REEL,
    "fldr": trace.Rev0.ORIG_FIELD_RECORD_NUM,
    "tracf": trace.Rev0.TRACE_NUM_ORIG_RECORD,
    "ep": trace.Rev0.ENERGY_SOURCE_POINT_NUM,
    "cdp": trace.Rev0.ENSEMBLE_NUM,
    "cdpt": trace.Rev0.TRACE_NUM_ENSEMBLE,
    "trid": trace.Rev0.TRACE_ID_CODE,
    "nvs": trace.Rev0.VERTICALLY_SUMMED_TRACES,
    "nhs": trace.Rev0.HORIZONTALLY_STACKED_TRACES,
    "duse": trace.Rev0.DATA_USE,
    "offset": trace.Rev0.SOURCE_TO_RECEIVER_DISTANCE,
    "gelev": trace.Rev0.RECEIVER_GROUP_ELEVATION,
    "selev": trace.Rev0.SOURCE_SURFACE_ELEVATION,
    "sdepth": trace.Rev0.SOURCE_DEPTH_BELOW_SURFACE,
    "gdel": trace.Rev0.RECEIVER_DATUM_ELEVATION,
    "sdel": trace.Rev0.SOURCE_DATUM_ELEVATION,
    "swdep": trace.Rev0.SOURCE_WATER_DEPTH,
    "gwdep": trace.Rev0.RECEIVER_WATER_DEPTH,
    "scalel": trace.Rev0.ELEVATION_DEPTH_SCALAR,
    "scalco": trace.Rev0.COORDINATE_SCALAR,
    "sx": trace.Rev0.SOURCE_COORD_X,
    "sy": trace.Rev0.SOURCE_COORD_Y,
    "gx": trace.Rev0.GROUP_COORD_X,
    "gy": trace.Rev0.GROUP_COORD_Y,
    "counit": trace.Rev0.COORDINATE_UNIT,
    "wevel": trace.Rev0.WEATHERING_VELOCITY,
    "swevel": trace.Rev0.SUBWEATHERING_VELOCITY,
    "sut": trace.Rev0.SOURCE_UPHOLE_TIME,
    "gut": trace.Rev0.GROUP_UPHOLE_TIME,
    "sstat": trace.Rev0.SOURCE_STATIC_CORRECTION,
    "gstat": trace.Rev0.RECEIVER_STATIC_CORRECTION,
    "tstat": trace.Rev0.TOTAL_STATIC_APPLIED,
    "laga": trace.Rev0.LAG_TIME_A,
    "lagb": trace.Rev0.LAG_TIME_B,
    "delrt": trace.Rev0.DELAY_RECORDING_TIME,
    "muts": trace.Rev0.MUTE_TIME_START,
    "mute": trace.Rev0.MUTE_TIME_END,
    "ns": trace.Rev0.SAMPLES_PER_TRACE,
    "dt": trace.Rev0.SAMPLE_INTERVAL,
    "gain": trace.Rev0.INSTRUMENT_GAIN_TYPE,
    "igc": trace.Rev0.INSTRUMENT_GAIN_CONST,
    "igi": trace.Rev0.INSTRUMENT_GAIN_INITIAL,
    "corr": trace.Rev0.CORRELATED_DATA,
    "sfs": trace.Rev0.SWEEP_FREQ_START,
    "sfe": trace.Rev0.SWEEP_FREQ_END,
    "slen": trace.Rev0.SWEEP_LENGTH,
    "styp": trace.Rev0.SWEEP_TYPE,
    "stas": trace.Rev0.SWEEP_TAPER_START,
    "stae": trace.Rev0.SWEEP_TAPER_END,
    "tatyp": trace.Rev0.TAPER_TYPE,
    "afilf": trace.Rev0.ALIAS_FILTER_FREQ,
    "afils": trace.Rev0.ALIAS_FILTER_SLOPE,
    "nofilf": trace.Rev0.NOTCH_FILTER_FREQ,
    "nofils": trace.Rev0.NOTCH_FILTER_SLOPE,
    "lcf": trace.Rev0.LOW_CUT_FREQ,
    "hcf": trace.Rev0.HIGH_CUT_FREQ,
    "lcs": trace.Rev0.LOW_CUT_SLOPE,
    "hcs": trace.Rev0.HIGH_CUT_SLOPE,
    "year": trace.Rev0.YEAR_RECORDED,
    "day": trace.Rev0.DAY_OF_YEAR,
    "hour": trace.Rev0.HOUR_OF_DAY,
    "minute": trace.Rev0.MINUTE_OF_HOUR,
    "sec": trace.Rev0.SECOND_OF_MINUTE,
    "timbas": trace.Rev0.TIME_BASIS_CODE,
    "trwf": trace.Rev0.TRACE_WEIGHTING_FACTOR,
    "grnors": trace.Rev0.GROUP_NUM_ROLL_SWITCH,
    "grnofr": trace.Rev0.GROUP_NUM_FIRST_TRACE,
    "grnlof": trace.Rev0.GROUP_NUM_LAST_TRACE,
    "gaps": trace.Rev0.GAP_SIZE,
    "otrav": trace.Rev0.TAPER_OVERTRAVEL,
}
# fmt: on



"""SEG-Y trace header definitions."""

from segy.standards.fields.base import SegStandardEnum


# fmt: off
class Rev0(SegStandardEnum):
    """Definition of SEG-Y Rev0 trace headers."""

    TRACE_SEQ_NUM_LINE          = (1, "int32")
    TRACE_SEQ_NUM_REEL          = (5, "int32")
    ORIG_FIELD_RECORD_NUM       = (9, "int32")
    TRACE_NUM_ORIG_RECORD       = (13, "int32")
    ENERGY_SOURCE_POINT_NUM     = (17, "int32")
    ENSEMBLE_NUM                = (21, "int32")
    TRACE_NUM_ENSEMBLE          = (25, "int32")
    TRACE_ID_CODE               = (29, "int16")
    VERTICALLY_SUMMED_TRACES    = (31, "int16")
    HORIZONTALLY_STACKED_TRACES = (33, "int16")
    DATA_USE                    = (35, "int16")
    SOURCE_TO_RECEIVER_DISTANCE = (37, "int32")
    RECEIVER_GROUP_ELEVATION    = (41, "int32")
    SOURCE_SURFACE_ELEVATION    = (45, "int32")
    SOURCE_DEPTH_BELOW_SURFACE  = (49, "int32")
    RECEIVER_DATUM_ELEVATION    = (53, "int32")
    SOURCE_DATUM_ELEVATION      = (57, "int32")
    SOURCE_WATER_DEPTH          = (61, "int32")
    RECEIVER_WATER_DEPTH        = (65, "int32")
    ELEVATION_DEPTH_SCALAR      = (69, "int16")
    COORDINATE_SCALAR           = (71, "int16")
    SOURCE_COORD_X              = (73, "int32")
    SOURCE_COORD_Y              = (77, "int32")
    GROUP_COORD_X               = (81, "int32")
    GROUP_COORD_Y               = (85, "int32")
    COORDINATE_UNIT             = (89, "int16")
    WEATHERING_VELOCITY         = (91, "int16")
    SUBWEATHERING_VELOCITY      = (93, "int16")
    SOURCE_UPHOLE_TIME          = (95, "int16")
    GROUP_UPHOLE_TIME           = (97, "int16")
    SOURCE_STATIC_CORRECTION    = (99, "int16")
    RECEIVER_STATIC_CORRECTION  = (101, "int16")
    TOTAL_STATIC_APPLIED        = (103, "int16")
    LAG_TIME_A                  = (105, "int16")
    LAG_TIME_B                  = (107, "int16")
    DELAY_RECORDING_TIME        = (109, "int16")
    MUTE_TIME_START             = (111, "int16")
    MUTE_TIME_END               = (113, "int16")
    SAMPLES_PER_TRACE           = (115, "int16")
    SAMPLE_INTERVAL             = (117, "int16")
    INSTRUMENT_GAIN_TYPE        = (119, "int16")
    INSTRUMENT_GAIN_CONST       = (121, "int16")
    INSTRUMENT_GAIN_INITIAL     = (123, "int16")
    CORRELATED_DATA             = (125, "int16")
    SWEEP_FREQ_START            = (127, "int16")
    SWEEP_FREQ_END              = (129, "int16")
    SWEEP_LENGTH                = (131, "int16")
    SWEEP_TYPE                  = (133, "int16")
    SWEEP_TAPER_START           = (135, "int16")
    SWEEP_TAPER_END             = (137, "int16")
    TAPER_TYPE                  = (139, "int16")
    ALIAS_FILTER_FREQ           = (141, "int16")
    ALIAS_FILTER_SLOPE          = (143, "int16")
    NOTCH_FILTER_FREQ           = (145, "int16")
    NOTCH_FILTER_SLOPE          = (147, "int16")
    LOW_CUT_FREQ                = (149, "int16")
    HIGH_CUT_FREQ               = (151, "int16")
    LOW_CUT_SLOPE               = (153, "int16")
    HIGH_CUT_SLOPE              = (155, "int16")
    YEAR_RECORDED               = (157, "int16")
    DAY_OF_YEAR                 = (159, "int16")
    HOUR_OF_DAY                 = (161, "int16")
    MINUTE_OF_HOUR              = (163, "int16")
    SECOND_OF_MINUTE            = (165, "int16")
    TIME_BASIS_CODE             = (167, "int16")
    TRACE_WEIGHTING_FACTOR      = (169, "int16")
    GROUP_NUM_ROLL_SWITCH       = (171, "int16")
    GROUP_NUM_FIRST_TRACE       = (173, "int16")
    GROUP_NUM_LAST_TRACE        = (175, "int16")
    GAP_SIZE                    = (177, "int16")
    TAPER_OVERTRAVEL            = (179, "int16")


class Rev1(SegStandardEnum):
    """Definition of SEG-Y Rev1 trace headers."""

    CDP_X                       = (181, "int32")
    CDP_Y                       = (185, "int32")
    INLINE                      = (189, "int32")
    CROSSLINE                   = (193, "int32")
    SHOT_POINT                  = (197, "int32")
    SHOT_POINT_SCALAR           = (201, "int16")
    TRACE_VALUE_UNIT            = (203, "int16")
    TRANSDUCTION_CONST_MANTISSA = (205, "int32")
    TRANSDUCTION_CONST_EXPONENT = (209, "int16")
    TRANSDUCTION_UNIT           = (211, "int16")
    DEVICE_TRACE_ID             = (213, "int16")
    TIMES_SCALAR                = (215, "int16")
    SOURCE_TYPE_ORIENTATION     = (217, "int16")
    SOURCE_ENERGY_DIR_MANTISSA  = (219, "int32")
    SOURCE_ENERGY_DIR_EXPONENT  = (223, "int16")
    SOURCE_MEASUREMENT_MANTISSA = (225, "int32")
    SOURCE_MEASUREMENT_EXPONENT = (229, "int16")
    SOURCE_MEASUREMENT_UNIT     = (231, "int16")


class Rev2(SegStandardEnum):
    """Definition of SEG-Y Rev2 trace headers."""

    TRACE_HEADER_NAME           = (233, "S8")
# fmt: on



"""Implementation of SEG-Y standards."""

from segy.schema.segy import SegyStandard
from segy.standards.registry import get_segy_standard
from segy.standards.registry import register_segy_standard
from segy.standards.spec import REV0
from segy.standards.spec import REV1
from segy.standards.spec import REV2
from segy.standards.spec import REV21

register_segy_standard(version_or_name=0.0, spec=REV0)
register_segy_standard(version_or_name=1.0, spec=REV1)
register_segy_standard(version_or_name=2.0, spec=REV2)
register_segy_standard(version_or_name=2.1, spec=REV21)


__all__ = ["get_segy_standard", "SegyStandard"]



"""Indexers for SEG-Y files."""

from __future__ import annotations

import logging
from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING
from typing import Any

import numpy as np
from fsspec.utils import merge_offset_ranges

from segy.arrays import HeaderArray
from segy.arrays import TraceArray
from segy.transforms import TransformPipeline

if TYPE_CHECKING:
    from pathlib import Path

    from fsspec import AbstractFileSystem
    from numpy.typing import NDArray

    from segy.schema import TraceSpec
    from segy.schema.base import BaseDataType

    IntDType = np.signedinteger[Any]


logger = logging.getLogger(__name__)


def merge_cat_file(
    fs: AbstractFileSystem,
    url: str | Path,
    starts: list[int],
    ends: list[int],
    block_size: int = 8_388_608,
) -> bytearray:
    """Merge sequential byte start/ends and fetch from store.

    Args:
        fs: fsspec FileSystem instance.
        url: Path/URL to file.
        starts: List of start byte locations.
        ends: List of end byte locations.
        block_size: Optional, block size for concurrent downloads.
            Default is 8MiB.

    Returns:
        Bytearray containing all the requested data.
    """
    paths = [url] * len(starts)

    paths, starts, ends = merge_offset_ranges(
        paths,
        starts,
        ends,
        max_block=block_size,
        sort=True,
    )

    buffer_bytes = fs.cat_ranges(
        paths=paths,
        starts=starts,
        ends=ends,
    )

    return bytearray(b"".join(buffer_bytes))


def bounds_check(indices: NDArray[IntDType], max_: int, type_: str) -> None:
    """Check if indices are out of bounds (negative, or more than max).

    Wrapping negative indices is not supported yet. The `type_` argument
    will be used in exceptions to be descriptive.

    Args:
        indices: A list of integer indices.
        max_: The maximum value of the index range.
        type_: The type of indices being checked.

    Raises:
        IndexError: If any of the indices are negative or exceed the maximum value.
    """
    oob_indices = np.where((indices < 0) | (indices >= max_))[0]

    if len(oob_indices) > 0:
        msg = (
            f"Requested {type_} indices {oob_indices} are out of bounds. SEG-Y "
            f"file has {max_} traces. Valid indices are "
            f"[0, {max_ - 1})."
        )
        raise IndexError(msg)


class AbstractIndexer(ABC):
    """Abstract class for indexing and fetching structured data from a remote file.

    We calculate byte ranges from indexing of SEG-Y components and use them
    to fetch the data and decode it.

    Args:
        fs: An instance of `fsspec` file-system.
        url: A string representing the URL of the file.
        spec: An instance of BaseDataType.
        max_value: An integer representing the maximum value of the index.
        transform_pipeline: The transforms pipeline to apply for decoding.

    Attributes:
        kind: A string representing the kind of data being indexed.
    """

    kind: str = "Abstract"

    def __init__(  # noqa: PLR0913
        self,
        fs: AbstractFileSystem,
        url: str,
        spec: BaseDataType,
        max_value: int,
        transform_pipeline: TransformPipeline | None = None,
    ):
        self.fs = fs
        self.url = url
        self.spec = spec
        self.max_value = max_value

        self.transform_pipeline = (
            TransformPipeline() if transform_pipeline is None else transform_pipeline
        )

    @abstractmethod
    def indices_to_byte_ranges(
        self, indices: NDArray[IntDType]
    ) -> tuple[NDArray[IntDType], NDArray[IntDType]]:
        """Logic to calculate start/end bytes."""

    @abstractmethod
    def decode(self, buffer: bytearray) -> NDArray[Any]:
        """How to decode the bytes after reading."""

    def post_process(self, data: NDArray[Any]) -> NDArray[Any]:
        """Apply transforms to the data after decoding."""
        return self.transform_pipeline.apply(data)

    def __getitem__(self, item: int | list[int] | NDArray[IntDType] | slice) -> Any:  # noqa: ANN401
        """Operator for integers, lists, and slices with bounds checking."""
        if isinstance(item, slice):
            if item.step == 0:
                msg = "Step of 0 is invalid for slicing."
                raise ValueError(msg)

            start = item.start or 0
            stop = item.stop or self.max_value
            start_stop = np.asarray([start, stop - 1])

            bounds_check(start_stop, self.max_value, self.kind)
            indices = np.arange(*item.indices(self.max_value))

        else:  # int, list, or ndarray case
            indices = np.atleast_1d(item)  # type: ignore
            bounds_check(indices, self.max_value, self.kind)

        if len(indices) == 0:
            msg = "Couldn't parse request. Please ensure it is a valid index."
            raise IndexError(msg)

        data = self.fetch(indices)
        return self.post_process(data)

    def fetch(self, indices: NDArray[IntDType]) -> NDArray[Any]:
        """Fetches and decodes binary data from the given indices.

        It supports duplicates in the indices, and it will also preserve
        the order of the request. If you want a sorted order, please sort
        the trace indices first.

        Args:
            indices: A list of integers representing the indices.

        Returns:
            An NDArray of any type representing the fetched data.

        Note:
            - This method internally converts the indices to byte ranges using
                the 'indices_to_byte_ranges' method.
            - The byte ranges are used to fetch the corresponding data from the
                file specified by the 'url' parameter. However, this is fastest
                if minimize the amount of reads. Here we combine starts and
                stops that are adjacent to each other. This requires a sort.
            - The fetched data is then decoded and squeezed before being returned.
        """
        unique_indices, index_order, counts = np.unique(
            indices,
            return_inverse=True,
            return_counts=True,
        )

        # Warn user about duplicates in the request
        if len(indices) != len(unique_indices):
            duplicate_mask = counts > 1
            values = unique_indices[duplicate_mask]
            counts = counts[duplicate_mask]
            duplicates = {int(v): int(c) for v, c in zip(values, counts, strict=True)}
            logger.warning("Duplicate indices requested with counts %s:", duplicates)

        starts, ends = self.indices_to_byte_ranges(indices)
        buffer = merge_cat_file(self.fs, self.url, starts.tolist(), ends.tolist())
        array = self.decode(buffer)
        return array[index_order].squeeze()


class TraceIndexer(AbstractIndexer):
    """Indexer for reading traces (headers + data).

    Inherits from AbstractIndexer. Implements decoding based on trace
    spec. It will optionally return the headers as a Pandas DataFrame.
    """

    spec: TraceSpec
    kind: str = "trace"

    def indices_to_byte_ranges(
        self, indices: NDArray[IntDType]
    ) -> tuple[NDArray[IntDType], NDArray[IntDType]]:
        """Convert trace indices to byte ranges."""
        if self.spec.offset is None:
            msg = "Trace starting offset must be specified."
            raise ValueError(msg)

        start_offset = self.spec.offset
        trace_itemsize = self.spec.dtype.itemsize

        starts = start_offset + indices * trace_itemsize
        ends = starts + trace_itemsize

        return starts, ends

    def decode(self, buffer: bytearray) -> TraceArray:
        """Decode whole traces (header + data)."""
        data = np.frombuffer(buffer, dtype=self.spec.dtype)
        return TraceArray(data)


class HeaderIndexer(AbstractIndexer):
    """Indexer for reading trace headers only.

    Inherits from AbstractIndexer. Implements decoding based on trace
    spec. It will optionally return the headers as a Pandas DataFrame.
    """

    spec: TraceSpec
    kind: str = "header"

    def indices_to_byte_ranges(
        self, indices: NDArray[IntDType]
    ) -> tuple[NDArray[IntDType], NDArray[IntDType]]:
        """Convert header indices to byte ranges (without trace data)."""
        trace_itemsize = self.spec.dtype.itemsize
        header_itemsize = self.spec.header.itemsize

        if self.spec.offset is None:
            msg = "Trace starting offset must be specified."
            raise ValueError(msg)

        start_offset = self.spec.offset

        starts = start_offset + indices * trace_itemsize
        ends = starts + header_itemsize

        return starts, ends

    def decode(self, buffer: bytearray) -> HeaderArray:
        """Decode headers only."""
        data = np.frombuffer(buffer, dtype=self.spec.dtype["header"])
        return HeaderArray(data)


class DataIndexer(AbstractIndexer):
    """Indexer for reading trace data samples only.

    Inherits from AbstractIndexer. Implements decoding based on trace spec.
    """

    spec: TraceSpec
    kind: str = "data"

    def indices_to_byte_ranges(
        self, indices: NDArray[IntDType]
    ) -> tuple[NDArray[IntDType], NDArray[IntDType]]:
        """Convert data indices to byte ranges (without trace headers)."""
        trace_itemsize = self.spec.itemsize
        data_itemsize = self.spec.data.itemsize
        header_itemsize = self.spec.header.itemsize

        if self.spec.offset is None:
            msg = "Trace starting offset must be specified."
            raise ValueError(msg)

        start_offset = self.spec.offset + header_itemsize

        starts = start_offset + indices * trace_itemsize
        ends = starts + data_itemsize

        return starts, ends

    def decode(self, buffer: bytearray) -> NDArray[Any]:
        """Decode trace samples only."""
        return np.frombuffer(buffer, dtype=self.spec.dtype["data"])



"""Data format specification representing scalar and trace data types."""

from __future__ import annotations

from enum import StrEnum
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from typing import Any


class ScalarType(StrEnum):
    """A class representing scalar data types."""

    IBM32 = "ibm32"
    INT64 = "int64"
    INT32 = "int32"
    INT16 = "int16"
    INT8 = "int8"
    UINT64 = "uint64"
    UINT32 = "uint32"
    UINT16 = "uint16"
    UINT8 = "uint8"
    FLOAT64 = "float64"
    FLOAT32 = "float32"
    FLOAT16 = "float16"
    STRING8 = "S8"

    def __repr__(self) -> str:
        """Nicer representation for users."""
        return f"{self.__class__.__name__}.{self._name_}"

    @property
    def dtype(self) -> np.dtype[Any]:
        """Return numpy dtype of the format."""
        # Special case for IBM 32-bit float
        if self.value == "ibm32":
            return np.dtype("uint32")

        return np.dtype(self.value)


class TextHeaderEncoding(StrEnum):
    """Supported textual header encodings."""

    ASCII = "ascii"
    EBCDIC = "ebcdic"

    @property
    def dtype(self) -> ScalarType:
        """Converts the byte order and data type of the object into a NumPy dtype."""
        return ScalarType.UINT8



"""Classes for managing headers and header groups."""

from __future__ import annotations

from functools import cached_property
from typing import Any

import numpy as np
from pydantic import Field

from segy.ebcdic import ASCII_TO_EBCDIC
from segy.ebcdic import EBCDIC_TO_ASCII
from segy.schema.base import BaseDataType
from segy.schema.format import TextHeaderEncoding  # noqa: TCH001


class TextProcessor:
    """Processor to transcode and (un)wrap text.

    Args:
        rows: Number of rows in text. Used for wrap/unwrap.
        cols: Number of columns in text. Used for wrap/unwrap.
        encoding: Text encoding used in transcoding.
    """

    def __init__(self, rows: int, cols: int, encoding: TextHeaderEncoding):
        self.rows = rows
        self.cols = cols
        self.encoding = encoding
        self.dtype = encoding.dtype

    def decode(self, buffer: bytes) -> str:
        """Decode bytes into a string given encoding."""
        if self.encoding == TextHeaderEncoding.EBCDIC:
            buffer_int = np.frombuffer(buffer, dtype=self.dtype)
            buffer = EBCDIC_TO_ASCII[buffer_int].tobytes()

        return buffer.decode("ascii", errors="replace")

    def encode(self, text: str) -> bytes:
        """Encode string into bytes given encoding."""
        buffer = text.encode("ascii")

        if self.encoding == TextHeaderEncoding.EBCDIC:
            buffer_int = np.frombuffer(buffer, dtype=self.dtype)
            buffer = ASCII_TO_EBCDIC[buffer_int].tobytes()

        return buffer

    def wrap(self, string: str) -> str:
        """Add line breaks after each row (n-cols)."""
        lines = []
        for row_idx in range(self.rows):
            start = row_idx * self.cols
            stop = start + self.cols
            lines.append(string[start:stop])

        return "\n".join(lines)

    @staticmethod
    def unwrap(text: str) -> str:
        """Remove all line breaks."""
        return text.replace("\n", "")


class TextHeaderSpec(BaseDataType):
    """A class representing spec for SEG-Y textual headers (8-bit)."""

    rows: int = Field(default=40, description="Number of rows in text header.")
    cols: int = Field(default=80, description="Number of columns in text header.")
    encoding: TextHeaderEncoding = Field(
        default=TextHeaderEncoding.EBCDIC, description="String encoding."
    )
    offset: int | None = Field(default=None, ge=0, description="Starting byte offset.")

    @cached_property
    def processor(self) -> TextProcessor:
        """Prepare transforms for encoding / decoding."""
        return TextProcessor(self.rows, self.cols, self.encoding)

    def __len__(self) -> int:
        """Get length of the textual header (number of characters)."""
        return self.rows * self.cols

    @property
    def dtype(self) -> np.dtype[Any]:
        """Get numpy dtype."""
        return np.dtype((self.encoding.dtype, len(self)))

    def decode(self, buffer: bytes) -> str:
        """Decode EBCDIC or ASCII bytes into string."""
        string = self.processor.decode(buffer)
        return self.processor.wrap(string)

    def encode(self, string: str) -> bytes:
        """Encode string to EBCDIC or ASCII bytes."""
        string = self.processor.unwrap(string)
        return self.processor.encode(string)


class ExtendedTextHeaderSpec(BaseDataType):
    """A class representing spec for SEG-Y extended textual headers."""

    spec: TextHeaderSpec = Field(..., description="Extended text header spec.")
    count: int = Field(default=0, ge=0, description="Number of extended text headers.")
    offset: int | None = Field(default=None, ge=0, description="Starting byte offset.")

    def __len__(self) -> int:
        """Get length of the textual header (number of characters)."""
        return len(self.spec) * self.count

    @property
    def dtype(self) -> np.dtype[Any]:
        """Get numpy dtype."""
        return np.dtype((self.spec.encoding.dtype, len(self)))

    def decode(self, buffer: bytes) -> list[str]:
        """Decode EBCDIC or ASCII bytes into string."""
        string = self.spec.processor.decode(buffer)

        chunk_size = len(self.spec)
        strings = []
        for start in range(0, len(self), chunk_size):
            stop = start + chunk_size
            text = self.spec.processor.wrap(string[start:stop])
            strings.append(text)

        return strings

    def encode(self, strings: list[str]) -> bytes:
        """Encode string to EBCDIC or ASCII bytes."""
        string = "".join(strings)
        string = self.spec.processor.unwrap(string)
        return self.spec.processor.encode(string)



"""Normalization logic for header field names."""

from rapidfuzz import process
from rapidfuzz.fuzz import WRatio

from segy.alias.segyio import SEGYIO_BIN_FIELD_MAP
from segy.alias.segyio import SEGYIO_TRACE_FIELD_MAP
from segy.alias.seis_unix import SEIS_UNIX_TRACE_FIELD_MAP
from segy.exceptions import InvalidFieldError
from segy.exceptions import NonSpecFieldError
from segy.standards.fields import binary
from segy.standards.fields import trace

FIELD_MAP = {}

FIELD_MAP.update({k.lower(): v.name.lower() for k, v in SEGYIO_BIN_FIELD_MAP.items()})
FIELD_MAP.update({k.lower(): v.name.lower() for k, v in SEGYIO_TRACE_FIELD_MAP.items()})
FIELD_MAP.update(
    {k.lower(): v.name.lower() for k, v in SEIS_UNIX_TRACE_FIELD_MAP.items()}
)

CANONICAL_KEYS = (
    set()
    # Binary header
    | {field.name for field in binary.Rev0}
    | {field.name for field in binary.Rev1}
    | {field.name for field in binary.Rev2}
    # Trace header
    | {field.name for field in trace.Rev0}
    | {field.name for field in trace.Rev1}
    | {field.name for field in trace.Rev2}
)
CANONICAL_KEYS = {name.lower() for name in CANONICAL_KEYS}


def process_str_for_fuzz(string: str) -> str:
    """Remove underscores and lowercase string."""
    return string.replace("_", "").lower()


def get_suggestion_keys(key: str) -> list[str]:
    """Get similar keys using weighted text similarity metrics."""
    suggestions = process.extract(
        key,
        CANONICAL_KEYS,
        limit=3,
        scorer=WRatio,
        processor=process_str_for_fuzz,
    )

    return [match[0] for match in suggestions]


def normalize_key(key: str) -> str:
    """Normalize a key to its canonical form."""
    key_lower = key.lower()
    if key_lower in CANONICAL_KEYS:
        return key_lower

    alias_key = FIELD_MAP.get(key_lower)

    if alias_key is None:
        suggestion_keys = get_suggestion_keys(key)
        raise InvalidFieldError(key, suggestion_keys)

    return alias_key


def validate_key(key: str, orig_key: str, defined_keys: tuple[str, ...]) -> None:
    """Validate a key against known fields and aliases."""
    if key not in defined_keys:
        msg = (
            f"The header field '{orig_key}' is found in field alias table as "
            f"'{key}'. However, the current SEG-Y spec does not define this "
            f"field in header fields."
        )
        raise NonSpecFieldError(msg)



"""Data model implementations for SEG-Y file spec."""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from pydantic import Field
from pydantic import model_validator

from segy.schema.base import CamelCaseModel

if TYPE_CHECKING:
    from segy.schema.base import Endianness
    from segy.schema.header import HeaderField
    from segy.schema.header import HeaderSpec
    from segy.schema.text_header import ExtendedTextHeaderSpec
    from segy.schema.text_header import TextHeaderSpec
    from segy.schema.trace import TraceDataSpec
    from segy.schema.trace import TraceSpec


class SegyStandard(float, Enum):
    """Allowed values for SEG-Y standards in SegySpec."""

    REV0 = 0.0
    REV1 = 1.0
    REV2 = 2.0
    REV21 = 2.1


class SegySpec(CamelCaseModel):
    """A class defining a SEG-Y file spec."""

    segy_standard: SegyStandard | None = Field(
        ..., description="SEG-Y Revision / Standard. Can also be custom."
    )
    text_header: TextHeaderSpec = Field(..., description="Textual file header spec.")
    binary_header: HeaderSpec = Field(..., description="Binary file header spec.")
    ext_text_header: ExtendedTextHeaderSpec | None = Field(
        default=None, description="Extended textual header spec."
    )
    trace: TraceSpec = Field(..., description="Trace header + data spec.")

    endianness: Endianness | None = Field(
        default=None, description="Endianness of SEG-Y file."
    )

    @model_validator(mode="after")
    def update_submodel_endianness(self) -> SegySpec:
        """Ensure that submodel endianness matches the SEG-Y endianness."""
        self.binary_header.endianness = self.endianness
        self.trace.endianness = self.endianness

        return self

    def update_offsets(self) -> None:
        """Update the offsets of the SEG-Y components."""
        cursor = 0

        if self.text_header.offset is None:
            self.text_header.offset = 0
        cursor += self.text_header.itemsize

        if self.binary_header.offset is None:
            self.binary_header.offset = cursor
        cursor += self.binary_header.itemsize

        if self.ext_text_header is not None:
            self.ext_text_header.offset = cursor
            cursor += self.ext_text_header.itemsize

        if self.trace.offset is None:
            self.trace.offset = cursor

    def customize(  # noqa: PLR0913
        self: SegySpec,
        text_header_spec: TextHeaderSpec | None = None,
        binary_header_fields: list[HeaderField] | None = None,
        ext_text_spec: ExtendedTextHeaderSpec | None = None,
        trace_header_fields: list[HeaderField] | None = None,
        trace_data_spec: TraceDataSpec | None = None,
    ) -> SegySpec:
        """Customize an existing SEG-Y spec.

        Args:
            text_header_spec: New text header specification.
            binary_header_fields: List of custom binary header fields.
            ext_text_spec: New extended text header spec.
            trace_header_fields: List of custom trace header fields.
            trace_data_spec: New trace data specification.

        Returns:
            A modified SEG-Y spec with "custom" segy standard.
        """
        new_spec = self.model_copy(deep=True)
        new_spec.segy_standard = None

        if text_header_spec is not None:
            new_spec.text_header = text_header_spec

        # Update binary header fields if specified; else will revert to default.
        if binary_header_fields is not None:
            new_spec.binary_header.customize(binary_header_fields)

        # Update extended text spec if its specified; else will revert to default.
        if ext_text_spec:
            new_spec.ext_text_header = ext_text_spec

        # Update trace header spec if its specified; else will revert to default.
        if trace_header_fields is not None:
            new_spec.trace.header.customize(trace_header_fields)

        # Update trace data spec if its specified; else will revert to default.
        if trace_data_spec:
            new_spec.trace.data = trace_data_spec

        return new_spec


class SegyInfo(CamelCaseModel):
    """Concise and useful information about SEG-Y files."""

    uri: str = Field(..., description="URI of the SEG-Y file.")

    segy_standard: SegyStandard | None = Field(
        ..., description="SEG-Y Revision / Standard. Can also be custom."
    )

    num_traces: int = Field(..., description="Number of traces.")

    samples_per_trace: int = Field(
        ..., description="Trace length in number of samples."
    )

    sample_interval: int | float = Field(
        ..., description="Sampling rate from binary header."
    )

    file_size: int = Field(..., description="File size in bytes.")



"""SEG-Y parser configuration."""

from __future__ import annotations

from typing import Any

from pydantic import Field
from pydantic_settings import BaseSettings
from pydantic_settings import SettingsConfigDict

from segy.schema import Endianness  # noqa: TCH001


class SegyBaseSettings(BaseSettings):
    """Base class for settings."""

    model_config = SettingsConfigDict(extra="ignore", case_sensitive=False)


class BinaryHeaderSettings(SegyBaseSettings):
    """SEG-Y binary header parsing overrides.

    Any value that is set to an integer will override what is parsed from
    the binary header in the actual file.
    """

    ext_text_header: int | None = Field(
        default=None, description="Override extended text headers."
    )
    revision: int | float | None = Field(
        default=None, description="SEG-Y revision of the file."
    )


class SegySettings(SegyBaseSettings):
    """SEG-Y file parsing settings."""

    binary: BinaryHeaderSettings = Field(
        default_factory=BinaryHeaderSettings,
        description="Overrides for binary file header settings.",
    )
    endianness: Endianness | None = Field(
        default=None,
        description="Override the inferred endianness of the file.",
    )
    storage_options: dict[str, Any] = Field(
        default_factory=dict,
        description="Storage options to pass to the storage backend.",
    )

    model_config = SettingsConfigDict(
        env_prefix="SEGY__",
        env_nested_delimiter="__",
    )



"""Custom array interface.

We subclass NumPy ndarray with some methods to enrich it for
better use experience. Like dictionary or JSON dumps for
structured arrays.

See here for details:
https://numpy.org/doc/stable/user/basics.subclassing.html
"""

from __future__ import annotations

from copy import copy
from json import dumps as json_dumps
from typing import TYPE_CHECKING
from typing import Any

import numpy as np

from segy.alias.core import normalize_key
from segy.alias.core import validate_key

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Iterable
    from collections.abc import Mapping
    from typing import Literal

    from numpy.typing import NDArray
    from pandas import DataFrame

    OrderKACF = Literal[None, "K", "A", "C", "F"]


class SegyArray(np.ndarray):  # type: ignore[type-arg]
    """Base class for array interface. Like ndarray but extensible."""

    def __new__(cls, input_array: NDArray[Any]) -> SegyArray:
        """Numpy subclass logic."""
        return np.asarray(input_array).view(cls)

    def __array_finalize__(self, obj: NDArray[Any] | None) -> None:
        """Numpy subclass logic."""
        if obj is None:
            return

    def __array_function__(
        self,
        func: Callable[..., Any],
        types: Iterable[type],
        args: Iterable[Any],
        kwargs: Mapping[str, Any],
    ) -> Any:  # noqa: ANN401
        """Ensure return type is still SegyArray and subtypes if we run numpy funcs.

        Functions like `np.concatenate` come here, and to ensure we keep the type
        as `SegyArray` or its subclasses when its run.
        """
        if not all(issubclass(t, SegyArray) for t in types):
            return NotImplemented

        result = super().__array_function__(func, types, args, kwargs)
        if isinstance(result, np.ndarray):
            return result.view(type(self))

        return result

    def copy(self, order: OrderKACF = "K") -> SegyArray:
        """Copy structured array preserving the padded bytes as is.

        This method ensures that the copy includes raw binary data and any padding
        bytes, preserving the entire memory layout of the array. This is necessary
        for working with SEG-Y data where not all fields are parsed, but raw binary
        data preservation is crucial.
        """
        void_view = self.view("V")
        void_copy = np.copy(void_view, order=order, subok=True)
        return void_copy.view(self.dtype)  # type: ignore[no-any-return]


class HeaderArray(SegyArray):
    """Header ndarray with convenience features."""

    def to_dict(self) -> dict[str, Any]:
        """Convert header to dict."""
        result_dict = {}

        if self.dtype.names is None:
            msg = f"{self.__class__.__name__} can only work on structured arrays."
            raise ValueError(msg)

        for field in self.dtype.names:
            field_values = self[field]
            result_dict[field] = field_values.squeeze().tolist()

        return result_dict

    def to_json(self, indent: int = 2) -> str:
        """Convert header to JSON."""
        return json_dumps(self.to_dict(), indent=indent)

    def to_dataframe(self) -> DataFrame:
        """Convert structured data to pandas DataFrame."""
        from pandas import DataFrame  # noqa: PLC0415 - Lazy import for fast module init

        return DataFrame.from_records(self)

    def _normalize_and_validate_keys(self, key: str | list[str]) -> str | list[str]:
        if isinstance(key, str):
            original_key = copy(key)
            key = normalize_key(key)
            validate_key(key, original_key, self.dtype.names)
        elif isinstance(key, list):
            original_keys = copy(key)
            key = [normalize_key(k) for k in key]
            for k, orig_k in zip(key, original_keys, strict=True):
                validate_key(k, orig_k, self.dtype.names)

        return key

    def __getitem__(self, item: Any) -> Any:  # noqa: ANN401
        """Special getitem where we normalize header keys. Pass along to numpy."""
        if isinstance(item, str) and item in self.dtype.names:
            return super().__getitem__(item)

        if isinstance(item, str):
            item = self._normalize_and_validate_keys(item)
        elif isinstance(item, list) and all(isinstance(i, str) for i in item):
            if all(key in self.dtype.names for key in item):
                return super().__getitem__(item)

            item = self._normalize_and_validate_keys(item)

        return super().__getitem__(item)

    def __setitem__(self, key: Any, value: Any) -> None:  # noqa: ANN401
        """Special getitem where we normalize header keys. Pass along to numpy."""
        if isinstance(key, str) and key in self.dtype.names:
            super().__setitem__(key, value)
            return

        if isinstance(key, str):
            key = self._normalize_and_validate_keys(key)

        elif isinstance(key, list) and all(isinstance(i, str) for i in key):
            if all(k in self.dtype.names for k in key):
                super().__setitem__(key, value)
                return

            key = self._normalize_and_validate_keys(key)

        super().__setitem__(key, value)


class TraceArray(SegyArray):
    """Trace ndarray with convenience features."""

    @property
    def header(self) -> HeaderArray:
        """Access headers of the trace(s)."""
        return HeaderArray(self["header"])

    @property
    def sample(self) -> NDArray[Any]:
        """Access data of the trace(s)."""
        return self["data"]



"""Factory methods for SEG-Y file creation."""

from __future__ import annotations

import logging
from datetime import UTC
from datetime import datetime
from typing import TYPE_CHECKING
from typing import cast

import numpy as np

from segy.arrays import HeaderArray
from segy.arrays import TraceArray
from segy.constants import REV1_BASE16
from segy.exceptions import NonSpecFieldError
from segy.schema.base import Endianness
from segy.schema.format import ScalarType
from segy.schema.segy import SegyStandard
from segy.standards.codes import DataSampleFormatCode
from segy.standards.codes import SegyEndianCode
from segy.transforms import TransformFactory
from segy.transforms import TransformPipeline

if TYPE_CHECKING:
    from typing import Any

    from numpy.typing import NDArray

    from segy.schema import SegySpec


logger = logging.getLogger(__name__)


DEFAULT_TEXT_HEADER_LINES = [
    "C01 File written by the open-source segy library.",
    "C02",
    "C03 Website: https://segy.readthedocs.io",
    "C04 Source: https://github.com/TGSAI/segy",
    "C05",
    "C06 File written: {timestamp}",
    "C07",
    "C08 SEG-Y Revision: {revision}",
    "C09",
    "C10 Sample Interval: {sample_interval} ms",
    "C11 Samples Per Trace: {samples_per_trace}",
]

DEFAULT_TEXT_HEADER_LINES += [f"C{line_no:02}" for line_no in range(12, 40)]
DEFAULT_TEXT_HEADER_LINES += ["C40 END TEXTUAL HEADER"]


def get_default_text(spec: SegySpec) -> str:
    """Dynamically generate default text header based on spec."""
    text_lines = DEFAULT_TEXT_HEADER_LINES.copy()

    # Populate write time
    now = datetime.now(tz=UTC).isoformat(timespec="seconds")
    text_lines[5] = text_lines[5].format(timestamp=now)

    # Populate revision
    revision = "unknown" if spec.segy_standard is None else spec.segy_standard.value
    text_lines[7] = text_lines[7].format(revision=revision)

    # Populate sample interval and number of samples
    sample_interval = spec.trace.data.interval // 1000  # type: ignore[operator]
    text_lines[9] = text_lines[9].format(sample_interval=sample_interval)
    text_lines[10] = text_lines[10].format(samples_per_trace=spec.trace.data.samples)

    text_lines = [line.ljust(80) for line in text_lines]
    return "\n".join(text_lines)


class SegyFactory:
    """Factory class for composing SEG-Y by components.

    Args:
        spec: SEG-Y specification.
        sample_interval: Sample interval to use in new file.
        samples_per_trace: Number of samples per trace.
    """

    def __init__(
        self,
        spec: SegySpec,
        sample_interval: int = 4000,
        samples_per_trace: int = 1500,
    ) -> None:
        self.spec = spec

        self.spec.trace.data.interval = sample_interval
        self.spec.trace.data.samples = samples_per_trace

    @property
    def sample_format(self) -> ScalarType:
        """Trace sample format of the SEG-Y file."""
        return self.spec.trace.data.format

    @property
    def sample_interval(self) -> int:
        """Return sample interval from spec."""
        # We know its populated at this point and its int (not None)
        return cast("int", self.spec.trace.data.interval)

    @property
    def samples_per_trace(self) -> int:
        """Return number of samples from spec."""
        # We know its populated at this point and its int (not None)
        return cast("int", self.spec.trace.data.samples)

    @property
    def segy_revision(self) -> SegyStandard:
        """Revision of the SEG-Y file."""
        if self.spec.segy_standard is None:
            return SegyStandard.REV0

        return self.spec.segy_standard

    def create_textual_header(self, text: str | None = None) -> bytes:
        """Create a textual header for the SEG-Y file.

        The length of the text should match the rows and columns in the spec's
        TextHeaderSpec. The newlines must also be in the text to separate
        the rows.

        Args:
            text: String containing text header rows. If left as None, a default
                textual header will be created.

        Returns:
            Bytes containing the encoded text header, ready to write.
        """
        if text is None:
            text = get_default_text(self.spec)

        text_spec = self.spec.text_header

        logger.info("Serialized text header according to SEG-Y spec.")
        return text_spec.encode(text)

    def create_binary_header(self, update: dict[str, Any] | None = None) -> bytes:
        """Create a binary header for the SEG-Y file.

        This function will create bytes representing the binary header with the
        configuration options provided to SEG-Y factory.

        The `update` parameter is a dictionary that contains binary header fields which
        need to be modified. The function will update these fields with the values
        specified in the update dictionary before returning the encoded bytes.

        Args:
            update: Dictionary containing binary header fields to modify.

        Returns:
            Bytes containing the encoded binary header, ready to write.
        """
        binary_spec = self.spec.binary_header
        bin_header = HeaderArray(np.zeros(shape=1, dtype=binary_spec.dtype))

        if self.segy_revision == SegyStandard.REV1:
            bin_header["segy_revision"] = REV1_BASE16
        elif self.segy_revision >= SegyStandard.REV2:
            minor, major = np.modf(self.segy_revision.value)
            minor = int(minor * 10)  # fraction to int
            bin_header["segy_revision_major"] = major
            bin_header["segy_revision_minor"] = minor
            bin_header["byte_order"] = SegyEndianCode.NATIVE

        bin_header["sample_interval"] = self.sample_interval
        bin_header["orig_sample_interval"] = self.sample_interval
        bin_header["samples_per_trace"] = self.samples_per_trace
        bin_header["orig_samples_per_trace"] = self.samples_per_trace
        bin_header["data_sample_format"] = DataSampleFormatCode[self.sample_format.name]

        if update is not None:
            for key, value in update.items():
                logger.debug("Updating binary header field %s to %s", key, value)
                bin_header[key] = value

        logger.info("Serialized binary header according to SEG-Y spec.")
        return bin_header.tobytes()

    def create_trace_header_template(
        self,
        size: int = 1,
    ) -> NDArray[Any]:
        """Create a trace header template array that conforms to the SEG-Y spec.

        Args:
            size: Number of headers for the template.

        Returns:
            Array containing the trace header template.
        """
        trace_header_spec = self.spec.trace.header
        dtype = trace_header_spec.dtype.newbyteorder("=")

        header_template = HeaderArray(np.zeros(shape=size, dtype=dtype))

        # 'names' assumed not None by data structure (type ignores).
        try:
            header_template["sample_interval"] = self.sample_interval
        except NonSpecFieldError:
            logger.warning("'sample_interval' not found in binary fields, skipping.")

        try:
            header_template["samples_per_trace"] = self.samples_per_trace
        except NonSpecFieldError:
            logger.warning("'samples_per_trace' not found in binary fields, skipping.")

        return header_template

    def create_trace_sample_template(
        self,
        size: int = 1,
    ) -> NDArray[Any]:
        """Create a trace data template array that conforms to the SEG-Y spec.

        Args:
            size: Number of traces for the template.

        Returns:
            Array containing the trace data template.
        """
        trace_data_spec = self.spec.trace.data
        dtype = trace_data_spec.dtype

        if self.sample_format == ScalarType.IBM32:
            dtype = np.dtype(("float32", (self.samples_per_trace,)))

        return np.zeros(shape=size, dtype=dtype)

    def create_traces(self, headers: NDArray[Any], samples: NDArray[Any]) -> bytes:
        """Convert trace data and header to bytes conforming to SEG-Y spec.

        The rows (length) of the headers and traces must match. The headers
        must be a (num_traces,) shape array and data must be a
        (num_traces, num_samples) shape array. They can be created via the
        `create_trace_header_template` and `create_trace_sample_template` methods.

        Args:
            headers: Header array.
            samples: Data array.

        Returns:
            Bytes containing the encoded traces, ready to write.

        Raises:
            AttributeError: if data dimensions are wrong (not 2D trace,samples).
            ValueError: if there is a shape mismatch between headers or num samples.
        """
        trace_spec = self.spec.trace
        trace_spec.data.samples = self.samples_per_trace

        if samples.ndim != 2:  # noqa: PLR2004
            msg = (
                "Data array must be 2-dimensional with rows as traces "
                "and columns as data samples."
            )
            logger.error(msg)
            raise AttributeError(msg)

        if samples.shape[1] != self.samples_per_trace:
            msg = f"Trace length must be {self.samples_per_trace}."
            logger.error(msg)
            raise ValueError(msg)

        if len(headers) != len(samples):
            msg = "Header array must have the same number of rows as data array."
            logger.error(msg)
            raise ValueError(msg)

        header_pipeline = TransformPipeline()
        data_pipeline = TransformPipeline()

        target_endian = trace_spec.endianness
        target_format = trace_spec.data.format

        if target_endian == Endianness.BIG:
            logger.debug("Added big endian conversion before serialization.")
            byte_swap = TransformFactory.create("byte_swap", target_endian)
            header_pipeline.add_transform(byte_swap)
            data_pipeline.add_transform(byte_swap)

        if target_format == ScalarType.IBM32:
            logger.debug("Added IBM float conversion before serialization.")
            ibm_float = TransformFactory.create("ibm_float", "to_ibm")
            data_pipeline.add_transform(ibm_float)

        trace = TraceArray(np.zeros(shape=headers.size, dtype=trace_spec.dtype))
        trace["header"] = header_pipeline.apply(headers)
        trace["data"] = data_pipeline.apply(samples)

        logger.info("Serialized %s traces according to SEG-Y spec.", len(samples))
        return trace.tobytes()



"""Accessors for accessing parts of TraceArrays."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from segy.schema import Endianness
from segy.schema import ScalarType
from segy.transforms import TransformFactory
from segy.transforms import TransformPipeline

if TYPE_CHECKING:
    from segy.schema import TraceSpec


logger = logging.getLogger(__name__)


class TraceAccessor:
    """Accessor for applying required transforms for reading SegyArrays and subclasses.

    trace_spec: Trace specification as instance of TraceSpec dtype.
    """

    def __init__(self, trace_spec: TraceSpec) -> None:
        logger.debug("Initializing %s", self.__class__.__name__)

        self.trace_spec = trace_spec
        self.header_ibm_keys = [
            field.name
            for field in self.trace_spec.header.fields
            if field.format == ScalarType.IBM32
        ]

        self.header_decode_pipeline = TransformPipeline()
        self.sample_decode_pipeline = TransformPipeline()
        self.trace_decode_pipeline = TransformPipeline()
        self._update_decoders()

    def _update_decoders(self) -> None:
        endian_transform = TransformFactory.create("byte_swap", Endianness.LITTLE)
        self.header_decode_pipeline.add_transform(endian_transform)
        self.sample_decode_pipeline.add_transform(endian_transform)

        if self.trace_spec.data.format == ScalarType.IBM32:
            ibm2ieee = TransformFactory.create("ibm_float", "to_ieee")
            self.sample_decode_pipeline.add_transform(ibm2ieee)

        if len(self.header_ibm_keys) != 0:
            ibm2ieee = TransformFactory.create(
                "ibm_float", "to_ieee", keys=self.header_ibm_keys
            )
            self.header_decode_pipeline.add_transform(ibm2ieee)

        trace_transform = TransformFactory.create(
            "trace", self.header_decode_pipeline, self.sample_decode_pipeline
        )
        self.trace_decode_pipeline.add_transform(trace_transform)



"""Command line interface for segy."""

import logging
import sys

import typer
from rich import print  # noqa: A004
from rich.logging import RichHandler

from segy.cli.common import JsonFileOutOption
from segy.cli.common import ListOfFieldNamesOption
from segy.cli.common import ListOfIntegersOption
from segy.cli.common import TextFileOutOption
from segy.cli.common import UriArgument
from segy.cli.common import modify_path
from segy.exceptions import InvalidFieldError
from segy.exceptions import NonSpecFieldError

FORMAT = "%(message)s"
logging.basicConfig(
    level=logging.ERROR,
    format=FORMAT,
    datefmt="[%X]",
    handlers=[RichHandler()],
)
logger = logging.getLogger("segy")

app = typer.Typer(
    name="dump",
    no_args_is_help=True,
    help="Dump SEG-Y file components to screen or file.",
)


@app.command(rich_help_panel="General Commands")
def info(uri: UriArgument, output: JsonFileOutOption = None) -> None:
    """Get basic information about a SEG-Y file."""
    from segy import SegyFile  # noqa: PLC0415 - Lazy import for fast init
    from segy.schema.segy import SegyInfo  # noqa: PLC0415 - Lazy import for fast init

    segy = SegyFile(uri)
    spec = segy.spec

    info = SegyInfo(
        uri=uri,
        segy_standard=spec.segy_standard,
        num_traces=segy.num_traces,
        samples_per_trace=segy.samples_per_trace,
        sample_interval=segy.sample_interval,
        file_size=segy.file_size,
    )

    info_json = info.model_dump_json(indent=2)

    if output is None:
        print(info_json)

    else:
        output = modify_path(output, suffix="info", default_extension=".json")
        with output.open(mode="w") as f:
            f.write(info_json)
        print(f"Saved SEG-Y info to file {output}")


@app.command(rich_help_panel="Header Commands")
def text_header(uri: UriArgument, output: TextFileOutOption = None) -> None:
    """Print or save the text header of a SEG-Y file."""
    from segy import SegyFile  # noqa: PLC0415 - Lazy import for fast init

    segy = SegyFile(uri)
    text = segy.text_header

    if output is None:
        print(text)

    else:
        output = modify_path(output, suffix="text", default_extension=".txt")
        with output.open(mode="w") as f:
            f.write(text)
        print(f"Saved SEG-Y text header to file {output}")


@app.command(rich_help_panel="Header Commands")
def binary_header(uri: UriArgument, output: JsonFileOutOption = None) -> None:
    """Print or save the binary header of a SEG-Y file."""
    from segy import SegyFile  # noqa: PLC0415 - Lazy import for fast init

    segy = SegyFile(uri)
    bin_header = segy.binary_header
    bin_header_json = bin_header.to_json()

    if output is None:
        print(bin_header_json)

    else:
        output = modify_path(output, suffix="binary", default_extension=".json")
        with output.open(mode="w") as f:
            f.write(bin_header_json)
        print(f"Saved SEG-Y binary header to file {output}")


@app.command(rich_help_panel="Trace Commands")
def trace(
    uri: UriArgument, index: ListOfIntegersOption, output: JsonFileOutOption = None
) -> None:
    """Get one or more traces with headers."""
    from segy import SegyFile  # noqa: PLC0415 - Lazy import for fast init

    segy = SegyFile(uri)

    print(segy.trace[index])

    if output is not None:
        msg = "Trace dump to file is not implemented yet."
        raise NotImplementedError(msg)


@app.command(rich_help_panel="Trace Commands")
def trace_header(
    uri: UriArgument,
    index: ListOfIntegersOption,
    field: ListOfFieldNamesOption,
    output: JsonFileOutOption = None,
) -> None:
    """Get one or more trace's headers (without samples)."""
    from pandas import Index  # noqa: PLC0415 - Lazy import for fast init

    from segy import SegyFile  # noqa: PLC0415 - Lazy import for fast init

    segy = SegyFile(uri)
    headers = segy.header[index]

    if len(field) > 0:
        try:
            headers = headers[field]
        except (InvalidFieldError, NonSpecFieldError) as e:
            logger.error(e)
            sys.exit(1)

    headers_df = headers.to_dataframe()

    row_index = Index(index, name="trace_index")
    headers_df.set_index(row_index, inplace=True)

    print(headers_df)

    if output is not None:
        msg = "Trace header dump to file is not implemented yet."
        raise NotImplementedError(msg)


@app.command(rich_help_panel="Trace Commands")
def trace_data(
    uri: UriArgument, index: ListOfIntegersOption, output: JsonFileOutOption = None
) -> None:
    """Get one or more trace's samples (without headers)."""
    from segy import SegyFile  # noqa: PLC0415 - Lazy import for fast init

    segy = SegyFile(uri)

    print(segy.sample[index])

    if output is not None:
        msg = "Trace data dump to file is not implemented yet."
        raise NotImplementedError(msg)



"""Specification representing header fields and headers."""

from __future__ import annotations

from collections import Counter
from typing import Any

import numpy as np
from pydantic import Field
from pydantic import field_validator
from pydantic import model_validator

from segy.schema.base import BaseDataType
from segy.schema.base import Endianness
from segy.schema.format import ScalarType  # noqa: TCH001


def _validate_non_overlapping_fields(fields: list[HeaderField]) -> None:
    """Validates that a list of HeaderField have unique names and do not overlap one-another.

    Args:
        fields: List of new header fields.

    Raises:
        ValueError: If duplicate header field names are detected.
        ValueError: If header fields overlap.
    """
    names = [field.name for field in fields]
    if len(names) != len(set(names)):
        msg = f"Duplicate header field names detected: {names}!"
        raise ValueError(msg)

    ranges = [field.range for field in fields]
    ranges.sort(key=lambda range_tuple: range_tuple[0])

    for i in range(len(ranges) - 1):
        if ranges_overlap(ranges[i], ranges[i + 1]):
            msg = f"Header fields overlap: {ranges[i]} and {ranges[i + 1]}!"
            raise ValueError(msg)


def ranges_overlap(range1: tuple[int, int], range2: tuple[int, int]) -> bool:
    """Checks if two right half-open ranges overlap."""
    return range1[0] < range2[1] and range1[1] > range2[0]


class HeaderField(BaseDataType):
    """A class representing header field spec.

    Examples:
        A named float starting at byte location 9:

        >>> field = HeaderField(
        >>>     name="my_var",
        >>>     format="float32",
        >>>     byte=9,
        >>> )

        The name, byte, and offset fields will only be used if the structured
        field is used within the context of a :class:`HeaderSpec`. Offset is
        calculated automatically from byte location.

        >>> field.name
        my_var
        >>> field.byte
        9
        >>> field.offset
        8

        The `dtype` property is inherited from :class:`DataFormat`.

        >>> field.dtype
        dtype('float32')
    """

    name: str = Field(..., description="The short name of the field.")
    byte: int = Field(..., ge=1, description="Field's start byte location.")
    format: ScalarType = Field(..., description="The data type of the field.")  # noqa: A003

    @property
    def offset(self) -> int:
        """Return zero based offset from one based byte location."""
        return self.byte - 1

    @property
    def dtype(self) -> np.dtype[Any]:
        """Converts the data type of the object into a NumPy dtype."""
        return self.format.dtype

    @property
    def range(self) -> tuple[int, int]:
        """Return the start and stop byte location of the field.

        Note: This return is Fortran-style and right half-open. [start, stop)
        """
        return self.byte, self.byte + self.dtype.itemsize


class HeaderSpec(BaseDataType):
    """A class representing a header specification.

    Examples:
        Let's build a header from scratch!

        We will define three fields with different names, data-types, and
        start byte locations.

        >>> field1 = HeaderField(
        >>>     name="foo",
        >>>     format="int32",
        >>>     byte=1,
        >>> )
        >>> field2 = HeaderField(
        >>>     name="bar",
        >>>     format="int16",
        >>>     byte=5,
        >>> )
        >>> field3 = HeaderField(
        >>>     name="fizz",
        >>>     format="int32",
        >>>     byte=17,
        >>> )

        Note that the fields span the following byte ranges:

        * `field1` between bytes `[0, 4)`
        * `field2` between bytes `[4, 6)`
        * `field3` between bytes `[16, 20)`

        The gap between `field2` and `field3` will be padded with `void`. In
        this case we expect to see an item size of 20-bytes (total length of
        the header struct).

        >>> header = HeaderSpec(
        >>>     fields=[field1, field2, field3],
        >>> )

        Now let's look at its data type:

        >>> header.dtype
        dtype({'names': ['foo', 'bar', 'fizz'], 'formats': ['<i4', '<i2', '<i4'], 'offsets': [0, 4, 16], 'itemsize': 20})

        If we wanted to pad the end of the struct (to fit a specific byte range),
        we would provide the item_size in the spec. If we set it to 30, this means
        that we padded the struct by 10 bytes at the end.

        >>> header = HeaderSpec(
        >>>     fields=[field1, field2, field3],
        >>>     item_size=30,
        >>> )

        Now let's look at its data type:

        >>> header.dtype
        dtype({'names': ['foo', 'bar', 'fizz'], 'formats': ['<i4', '<i2', '<i4'], 'offsets': [0, 4, 16], 'itemsize': 30})

        To see what's going under the hood, we can look at a lower level numpy
        description of the `dtype`. Here we observe all the gaps (void types).

        >>> header.dtype.descr
        [('foo', '<i4'), ('bar', '<i2'), ('', '|V10'), ('fizz', '<i4'), ('', '|V10')]
    """  # noqa: E501

    fields: list[HeaderField] = Field(
        ..., description="List containing multiple header field spec instances."
    )
    item_size: int | None = Field(
        default=None, description="Expected size of the struct."
    )
    offset: int | None = Field(default=None, ge=0, description="Starting byte offset.")

    endianness: Endianness | None = Field(
        default=None, description="Endianness of structured data type."
    )

    @property
    def dtype(self) -> np.dtype[Any]:
        """Converts the names, data types, and offsets of the object into a NumPy dtype."""
        dtype_conf = {
            "names": self.names,
            "formats": self.formats,
            "offsets": self.offsets,
        }

        if self.item_size is not None:
            dtype_conf["itemsize"] = self.item_size

        struct_dtype = np.dtype(dtype_conf)  # type: ignore[call-overload]

        if self.endianness is not None:
            struct_dtype = struct_dtype.newbyteorder(self.endianness.symbol)

        return struct_dtype  # type: ignore[no-any-return]

    @property
    def names(self) -> list[str]:
        """Get the names of the fields."""
        return [field.name for field in self.fields]

    @property
    def formats(self) -> list[np.dtype[Any]]:
        """Get the formats of the fields."""
        return [field.dtype for field in self.fields]

    @property
    def offsets(self) -> list[int]:
        """Get the offsets the fields."""
        return [field.offset for field in self.fields]

    @field_validator("fields")
    @classmethod
    def ensure_no_duplicate_fields(
        cls: type[HeaderSpec], fields: list[HeaderField]
    ) -> list[HeaderField]:
        """Check if fields are unique and error out if not."""
        name_counter = Counter(field.name for field in fields)
        duplicates = [name for name, count in name_counter.items() if count > 1]

        if duplicates:
            msg = f"Duplicate header fields detected: {', '.join(duplicates)}."
            raise ValueError(msg)

        return fields

    @model_validator(mode="after")
    def ensure_offsets_in_itemsize_bounds(self) -> HeaderSpec:
        """Ensure fields don't go above the allowed itemsize."""
        if len(self.fields) == 0:
            return self

        max_field = max(self.fields, key=lambda field: field.offset)

        if self.item_size is None:
            return self

        if max_field.offset + max_field.dtype.itemsize > self.item_size:
            msg = "Offsets exceed allowed header size."
            raise ValueError(msg)

        return self

    def add_field(self, field: HeaderField, overwrite: bool = False) -> None:
        """Add a field to the structured data type."""
        if field.name in self.names and overwrite is False:
            msg = (
                f"Field named {field.name} already exists. If you wish to "
                "overwrite an existing field, pass `overwrite=True`."
            )
            raise KeyError(msg)

        if field.name not in self.names:
            self.fields.append(field)
        else:
            field_idx = self.names.index(field.name)
            self.fields[field_idx] = field

        # Trigger validation
        self.fields = self.fields

    def remove_field(self, name: str) -> None:
        """Remove a field from the structured data type by name."""
        try:
            field_idx = self.names.index(name)
        except ValueError as err:
            msg = f"Field named {name} does not exist."
            raise KeyError(msg) from err

        del self.fields[field_idx]

    def customize(self, fields: HeaderField | list[HeaderField]) -> None:
        """Customizes existing HeaderSpec fields with new headers handling overlaps.

        It first handles name conflicts. Then it handles byte-range intersections.

        This ensures no byte-range intersections between new and existing fields. Assumes
        new fields do not overlap each other (validated first) and existing fields do not
        overlap each other.

        Args:
            fields: List of new header fields.
        """
        if isinstance(fields, HeaderField):
            fields = [fields]

        _validate_non_overlapping_fields(fields)

        # Handle name conflicts (overwrite old ones)
        for field in fields:
            self.add_field(field, overwrite=True)

        # Handle byte-range conflicts (remove overlapping old fields)
        fields_to_remove = set()
        for new_field in fields:
            for field in self.fields:
                does_overlap = ranges_overlap(field.range, new_field.range)
                if field.name != new_field.name and does_overlap:
                    fields_to_remove.add(field.name)

        for name in fields_to_remove:
            self.remove_field(name)



"""Utilities to extract information from the SEG-Y files.

We have the following inference options:
1. Endianness inference
2. Revision interpretation
"""

from __future__ import annotations

import logging
import sys
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING

import numpy as np

from segy.config import SegySettings
from segy.exceptions import EndiannessInferenceError
from segy.schema import Endianness
from segy.schema import SegyStandard
from segy.standards.codes import DataSampleFormatCode
from segy.standards.codes import SegyEndianCode

if TYPE_CHECKING:
    from numpy._typing import DTypeLike

logger = logging.getLogger(__name__)


class EndiannessAction(Enum):
    """Descriptive flag enum for endianness reversal."""

    REVERSE = True
    KEEP = False


@dataclass
class SegyInferResult:
    """A scan result of a SEG-Y file.

    Attributes:
        endianness: Endianness of the file.
        revision: SEG-Y revision as float.
    """

    __slots__ = ("endianness", "revision")

    endianness: Endianness
    revision: float


def infer_endianness(
    buffer: bytes,
    settings: SegySettings,
) -> EndiannessAction:
    """Infer if we need to reverse the endianness of the seismic data.

    Args:
        buffer: Bytes representing the binary header.
        settings: Settings instance to configure / override.

    Returns:
        A boolean indicating if the endianness need to be reversed.

    Raises:
        EndiannessInferenceError: When all methods fail.
        NotImplementedError: When a pairwise swapped endianness is detected.
    """
    logger.debug("Starting endianness inference.")

    # Method 1: Use settings if available
    if settings.endianness is not None:
        logger.info("Using provided endianness from settings: %s", settings.endianness)
        return EndiannessAction(settings.endianness != sys.byteorder)

    # Method 2: Explicit endianness code (SEG-Y Rev2+)
    logger.debug("Trying explicit endianness code (SEGY Rev2+).")
    endian_code = np.frombuffer(buffer, "uint32", offset=96, count=1)[0]

    if endian_code == SegyEndianCode.NATIVE:
        logger.info("Detected native endianness.")
        return EndiannessAction.KEEP
    if endian_code == SegyEndianCode.REVERSE:
        logger.info("Detected reverse endianness.")
        return EndiannessAction.REVERSE
    if endian_code == SegyEndianCode.PAIRWISE_SWAP:
        msg = "Pairwise swapped endianness detected. Not supported."
        logger.error(msg)
        raise NotImplementedError(msg)
    if endian_code != 0:
        logger.warning("Ambiguous explicit endianness code: %s", endian_code)

    # Method 3: Legacy method using sample format for inference (SEG-Y <Rev2)
    logger.debug("Trying legacy method for SEGY Rev <2.0.")
    format_dtype = np.dtype("uint16")
    supported_formats = set(DataSampleFormatCode._value2member_map_.keys())

    def _is_supported_format(dtype: DTypeLike) -> bool:
        format_value = np.frombuffer(buffer, dtype, offset=24, count=1)[0]
        return format_value in supported_formats

    if _is_supported_format(format_dtype):
        logger.info("Detected native endianness using legacy method.")
        return EndiannessAction.KEEP

    if _is_supported_format(format_dtype.newbyteorder()):
        logger.info("Detected reverse endianness using legacy method.")
        return EndiannessAction.REVERSE

    # If all methods fail
    error_message = (
        "Endianness inference failed after attempting all methods. "
        "Ensure the file is valid or provide explicit settings."
    )
    logger.error(error_message)
    raise EndiannessInferenceError(error_message)


def interpret_revision(
    buffer: bytes,
    endianness_action: EndiannessAction | None = None,
    settings: SegySettings | None = None,
) -> int | float:
    """Infer the revision number from the binary header of a SEG-Y file.

    Args:
        buffer: The binary header buffer.
        endianness_action: The action to take for endianness.
        settings: The SegySettings, which may override the revision.

    Returns:
        The revision number as a float (e.g., 1.0, 1.2, 2.0).
    """
    if endianness_action is None:
        endianness_action = EndiannessAction.KEEP

    if settings is None:
        settings = SegySettings()

    logger.debug("Starting revision inference.")

    # Method 1: Use settings if available
    if settings.binary.revision is not None:
        settings_rev = settings.binary.revision
        logger.info("Using provided revision from settings: %s", settings_rev)
        return settings_rev

    # Method 2: Major/minor from single byte integers (SEG-Y Rev2+)
    logger.debug("Checking if file is SEG-Y Rev2+.")
    major_revision = np.frombuffer(buffer, "uint8", offset=300, count=1)[0]

    if major_revision >= SegyStandard.REV2:
        minor_revision = np.frombuffer(buffer, "uint8", offset=301, count=1)[0]
    else:
        # Method 3: Major/minor from 16-bit integer (SEG-Y <Rev2)
        logger.debug("File is SEG-Y <Rev2, reading revision from 16-bits.")
        dtype = np.dtype("uint16")
        if endianness_action == EndiannessAction.REVERSE:
            dtype = dtype.newbyteorder()

        revision = np.frombuffer(buffer, dtype, offset=300, count=1)[0]
        major_revision = revision >> 8
        minor_revision = revision & 0xFF

    revision_float = int(major_revision) + int(minor_revision) / 10
    logger.info("Detected revision from binary header as %s", revision_float)
    return revision_float



"""Implements a registry for various SEG-Y standards."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from segy.schema import SegySpec


logger = logging.getLogger(__name__)


segy_standard_registry = {}


def register_segy_standard(version_or_name: float | str, spec: SegySpec) -> None:
    """Register a SEG-Y standard with its spec."""
    logger.info("Registered new SEG-Y spec: %s", version_or_name)
    segy_standard_registry[version_or_name] = spec


def get_segy_standard(version_or_name: float | str) -> SegySpec:
    """Get a registered SEG-Y standard's spec."""
    spec = segy_standard_registry.get(version_or_name)

    if spec is None:
        msg = (
            f"Unknown or unsupported SEG-Y spec: {version_or_name}. If you "
            f"would like to use {version_or_name}, please register it with "
            f"the `SegySpecFactory` using its `register_spec` method."
        )
        logger.error(msg)
        raise NotImplementedError(msg)

    return spec.model_copy(deep=True)



"""Transforms to apply to arrays and structured arrays."""

from __future__ import annotations

import logging
import sys
from abc import abstractmethod
from typing import TYPE_CHECKING

import numpy as np
from numpy.lib import recfunctions as rfn

from segy.inference import interpret_revision
from segy.schema import SegyStandard
from segy.schema.base import Endianness

if TYPE_CHECKING:
    from typing import Any

    from numpy._typing._dtype_like import _DTypeDict
    from numpy.typing import DTypeLike
    from numpy.typing import NDArray


logger = logging.getLogger(__name__)


def get_endianness(data: NDArray[Any]) -> Endianness:
    """Map the numpy byte order character to Endianness."""
    if data.dtype.isnative:
        return Endianness[sys.byteorder.upper()]

    return Endianness.LITTLE if data.dtype.byteorder == "<" else Endianness.BIG


def _extract_nested_dtype(dtype: np.dtype[Any]) -> DTypeLike:
    """Extract nested dtype information from struct dtype."""
    if dtype.names is None or dtype.fields is None:
        return dtype

    dtype_info: _DTypeDict = {
        "names": dtype.names,
        "formats": [
            _extract_nested_dtype(dtype.fields[name][0]) for name in dtype.names
        ],
        "offsets": [dtype.fields[name][1] for name in dtype.names],
        "itemsize": dtype.itemsize,
    }

    return dtype_info


def _modify_dtype_field(
    old_dtype: np.dtype[np.void],
    key: str,
    new_type: np.dtype[Any],
) -> np.dtype[np.void]:
    """Constructs a new dtype for the structured array after a type change in one of its fields."""
    if old_dtype.names is None:  # pragma: no cover
        msg = "Cannot modify dtype for non-structured array."
        raise ValueError(msg)

    dtype_info: _DTypeDict = _extract_nested_dtype(old_dtype)  # type: ignore[assignment]

    key_index = dtype_info["names"].index(key)
    key_to_modify = dtype_info["formats"][key_index]

    # Handle the case where field to modify is a struct
    # For example; a Trace struct with "header" amd "data" fields.
    if isinstance(key_to_modify, np.dtype) and key_to_modify.kind == "V":
        new_type = np.dtype((new_type.str,) + key_to_modify.subdtype[1:])  # type: ignore

    dtype_info["formats"][key_index] = new_type  # type: ignore[index]

    return np.dtype(dtype_info)


def _modify_structured_field(
    data: NDArray[np.void],
    key: str,
    transformed: NDArray[Any],
) -> NDArray[np.void]:
    """Handle structured array assignment with dtype change or promotion."""
    if data.dtype.names is None:  # pragma: no cover
        msg = "Cannot modify dtype for non-structured array."
        raise ValueError(msg)

    # Assign and return early if basic types are the same. Numpy handles endianness.
    if transformed.dtype == data[key].dtype:
        data[key] = transformed
        return data

    # If transform changed field type, we need to define the new array dtype.
    old_field_dtype = data[key].dtype
    new_field_dtype = transformed.dtype
    new_struct_dtype = _modify_dtype_field(data.dtype, key, new_field_dtype)

    # If the field size hasn't changed we can do it in-place with view magic, no copy.
    # Example: ibm32 to float32 still fits in the same 32-bits.
    data[key] = transformed.view(old_field_dtype)
    return data.view(new_struct_dtype)


class Transform:
    """Base class for header transformation strategies."""

    def __init__(self, keys: list[str] | None = None, copy: bool = False) -> None:
        self.keys = keys
        self.copy = copy

        logger.debug("Initialized %s.", self.__class__.__name__)

    def apply(self, data: NDArray[Any]) -> NDArray[Any]:
        """Applies transformation based on ndarray or struct array.

        For memory efficiency, by default, the transforms are applied to input array.
        This could cause mutation of input array depending on the transform. Set the
        `copy` option to `True` if you want to ensure input data isn't modified.

        Args:
            data: The array to be transformed.

        Returns:
            Modified array or copy of array.
        """
        if self.copy:
            data = data.copy()

        if self.keys is None:
            return self.transform(data)

        return self.transform_struct(data)

    @abstractmethod
    def transform(self, data: NDArray[Any]) -> NDArray[Any]:
        """Abstract transformation method on numpy arrays."""

    def transform_struct(self, data: NDArray[Any]) -> NDArray[Any]:
        """Generalized transform for structured arrays."""
        if self.keys is None:  # pragma: no cover
            msg = "Trying to modify structured array fields with no keys provided."
            raise ValueError(msg)

        if data.dtype.names is None:
            msg = "The array to transform doesn't contain any fields."
            raise ValueError(msg)

        missing_fields = set(self.keys) - set(data.dtype.names)
        if missing_fields:
            msg = f"Field(s) {missing_fields} not in the array."
            raise ValueError(msg)

        for key in self.keys:
            transformed = self.transform(data[key])
            data = _modify_structured_field(data, key, transformed)

        return data


class ByteSwapTransform(Transform):
    """Byte swaps numeric data by based on target order.

    Args:
        target_order: Desired byte order.
    """

    def __init__(self, target_order: Endianness) -> None:
        super().__init__()
        self.target_order = target_order
        logger.debug("Converting to %s endian.", self.target_order.value)

    def transform(self, data: NDArray[Any]) -> NDArray[Any]:
        """Byte swap numpy array given target order."""
        source_order = get_endianness(data)

        if source_order != self.target_order:
            data = data.byteswap(inplace=True)
            swapped_dtype = data.dtype.newbyteorder(self.target_order.symbol)
            data = data.view(swapped_dtype)

        return data


class IbmFloatTransform(Transform):
    """IBM float convert an array.

    Args:
        direction: IBM Float conversion direction.
        keys: Optional list of keys to apply the transform.

    Attributes:
        ibm_func_map: Mapping of ibm conversion direction and function.
    """

    # To map user parameter to compiled function and its expected type.
    ibm_func_map = {
        "to_ibm": ("ieee2ibm", "float32"),
        "to_ieee": ("ibm2ieee", "uint32"),
    }

    def __init__(self, direction: str, keys: list[str] | None = None) -> None:
        super().__init__(keys)
        self.direction = direction
        logger.debug("Converting %s.", self.direction)

    def transform(self, data: NDArray[Any]) -> NDArray[Any]:
        """Convert floats between IEEE and IBM."""
        from segy import ibm  # noqa: PLC0415 - Lazy import for fast module init

        func_name, cast_dtype = self.ibm_func_map[self.direction]
        func = getattr(ibm, func_name)

        return func(data.astype(cast_dtype))  # type: ignore


class SegyRevisionTransform(Transform):
    """Interpret the SEG-Y revision field in binary header."""

    def __init__(self) -> None:
        super().__init__()

    def transform(self, data: NDArray[Any]) -> NDArray[Any]:
        """Parse SEG-Y standard from binary header."""
        revision = interpret_revision(data.tobytes())
        minor, major = np.modf(revision)
        minor = int(minor * 10)  # fraction to int

        # Well defined in Rev2+, don't modify
        if revision >= SegyStandard.REV2:
            return data

        # Normalize it all to rev2+
        # It is ambiguous due to SEG-Y and segyio conventions
        data = rfn.drop_fields(
            data,
            drop_names=["segy_revision", "segy_revision_major", "segy_revision_minor"],
            usemask=False,
        )

        return rfn.append_fields(  # type: ignore
            data,
            names=["segy_revision_major", "segy_revision_minor"],
            data=[major, minor],
            dtypes=["uint8", "uint8"],
            usemask=False,
        )


class TraceTransform(Transform):
    """Composite transform to apply header and data pipeline to trace.

    This transform is a workaround for a design flaw in applying transforms.
    Please refactor this at some point! The problem is: if we have a trace
    struct with "headers" and "data", and we want to apply another transform
    to a header field within "headers", we can't. This is a problem with the
    current Transform and TransformPipeline.

    Args:
        header_pipeline: Transform pipeline to apply to struct "header" field.
        data_pipeline: Transform pipeline to apply to struct "header" field.
    """

    def __init__(
        self, header_pipeline: TransformPipeline, data_pipeline: TransformPipeline
    ) -> None:
        super().__init__()
        self.header_transform = Transform(["header"])  # type: ignore[abstract]
        self.header_transform.transform = header_pipeline.apply  # type: ignore[method-assign]

        self.data_transform = Transform(["data"])  # type: ignore[abstract]
        self.data_transform.transform = data_pipeline.apply  # type: ignore[method-assign]

    def transform(self, data: NDArray[Any]) -> NDArray[Any]:
        """Applies independent transform pipelines to trace struct."""
        data = self.header_transform.apply(data)
        data = self.data_transform.apply(data)

        return data  # noqa: RET504


class TransformFactory:
    """Factory class to generate transformation strategies."""

    transform_map: dict[str, type[Transform]] = {
        "byte_swap": ByteSwapTransform,
        "ibm_float": IbmFloatTransform,
        "segy_revision": SegyRevisionTransform,
        "trace": TraceTransform,
    }

    @classmethod
    def create(
        cls: type[TransformFactory],
        transform_type: str,
        *args: Any,  # noqa: ANN401
        **kwargs: Any,  # noqa: ANN401
    ) -> Transform:
        """Create an instance of transformation and return it."""
        if transform_type not in cls.transform_map:
            msg = f"Unsupported transformation type: {transform_type}"
            logger.error(msg)
            raise KeyError(msg)

        return cls.transform_map[transform_type](*args, **kwargs)


class TransformPipeline:
    """Executes a sequence of transformations on data."""

    def __init__(self) -> None:
        self.transforms: list[Transform] = []

    def add_transform(self, transform: Transform) -> None:
        """Adds a transformation to the pipeline."""
        self.transforms.append(transform)

    def apply(self, data: NDArray[Any]) -> NDArray[Any]:
        """Applies all transformations in sequence."""
        for transform in self.transforms:
            data = transform.apply(data)
        return data



"""Data models for manipulating all kinds of SEG-Y files."""

from segy.schema.base import Endianness
from segy.schema.format import ScalarType
from segy.schema.format import TextHeaderEncoding
from segy.schema.header import HeaderField
from segy.schema.header import HeaderSpec
from segy.schema.segy import SegySpec
from segy.schema.segy import SegyStandard
from segy.schema.text_header import ExtendedTextHeaderSpec
from segy.schema.text_header import TextHeaderSpec
from segy.schema.trace import TraceDataSpec
from segy.schema.trace import TraceSpec

__all__ = [
    "Endianness",
    "ExtendedTextHeaderSpec",
    "ScalarType",
    "HeaderField",
    "HeaderSpec",
    "SegySpec",
    "SegyStandard",
    "TextHeaderEncoding",
    "TextHeaderSpec",
    "TraceDataSpec",
    "TraceSpec",
]



"""SEG-Y file in-memory constructors."""

from __future__ import annotations

import logging
import sys
from functools import cached_property
from typing import TYPE_CHECKING
from typing import cast

import numpy as np
from fsspec.core import url_to_fs

from segy.accessors import TraceAccessor
from segy.arrays import HeaderArray
from segy.config import SegySettings
from segy.exceptions import SegyFileSpecMismatchError
from segy.indexing import DataIndexer
from segy.indexing import HeaderIndexer
from segy.indexing import TraceIndexer
from segy.inference import EndiannessAction
from segy.inference import SegyInferResult
from segy.inference import infer_endianness
from segy.inference import interpret_revision
from segy.schema import Endianness
from segy.schema import ScalarType
from segy.standards import get_segy_standard
from segy.standards.codes import DataSampleFormatCode
from segy.transforms import TransformFactory
from segy.transforms import TransformPipeline

if TYPE_CHECKING:
    from fsspec import AbstractFileSystem
    from numpy.typing import NDArray

    from segy.indexing import AbstractIndexer
    from segy.schema import SegySpec

logger = logging.getLogger(__name__)


class SegyFile:
    """A SEG-Y file class that has various accessors.

    Args:
        url: Path to SEG-Y file on disk or remote store.
        spec: The schema / spec describing the SEG-Y file. This
            is optional and by default it will try to infer the
            SEG-Y standard from the binary header.
        settings: A settings instance to configure / override
            the SEG-Y parsing logic. Optional.

    Attributes:
        fs: The filesystem instance used to access the file.
        url: The URL of the SEG-Y file on the filesystem.
    """

    fs: AbstractFileSystem
    url: str

    def __init__(
        self,
        url: str,
        spec: SegySpec | None = None,
        settings: SegySettings | None = None,
    ):
        self.settings = settings if settings is not None else SegySettings()

        self.fs, self.url = url_to_fs(url, **self.settings.storage_options)
        self._info = self.fs.info(self.url)

        self.spec = self._initialize_spec(spec)
        self._update_spec()

        self.accessors = TraceAccessor(self.spec.trace)

    @property
    def file_size(self) -> int:
        """Return file size in bytes."""
        size: int = self._info["size"]
        return size

    @property
    def samples_per_trace(self) -> int:
        """Return samples per trace in file based on spec."""
        return cast("int", self.spec.trace.data.samples)  # we know for sure its int

    @property
    def sample_interval(self) -> int:
        """Return samples interval in file based on spec."""
        return cast("int", self.spec.trace.data.interval)  # we know for sure its int

    @property
    def sample_labels(self) -> NDArray[np.int32]:
        """Return sample axis labels."""
        max_samp = self.sample_interval * self.samples_per_trace
        return np.arange(0, max_samp, self.sample_interval, dtype="int32")

    @property
    def num_ext_text(self) -> int:
        """Return number of extended text headers."""
        if self.spec.ext_text_header is None:
            return 0

        return self.spec.ext_text_header.count

    @property
    def num_traces(self) -> int:
        """Return number of traces in file based on size and spec."""
        return cast("int", self.spec.trace.count)  # we know for sure its int

    @cached_property
    def text_header(self) -> str:
        """Return textual file header."""
        text_hdr_spec = self.spec.text_header

        buffer = self.fs.read_block(
            fn=self.url,
            offset=text_hdr_spec.offset,
            length=text_hdr_spec.itemsize,
        )

        return text_hdr_spec.decode(buffer)

    @cached_property
    def ext_text_header(self) -> list[str]:
        """Return textual file header."""
        ext_text_hdr_spec = self.spec.ext_text_header

        if ext_text_hdr_spec is None:
            return []

        buffer = self.fs.read_block(
            fn=self.url,
            offset=ext_text_hdr_spec.offset,
            length=ext_text_hdr_spec.itemsize,
        )

        return ext_text_hdr_spec.decode(buffer)

    @cached_property
    def binary_header(self) -> HeaderArray:
        """Read binary header from store, based on spec."""
        buffer_bytes = self.fs.read_block(
            fn=self.url,
            offset=self.spec.binary_header.offset,
            length=self.spec.binary_header.itemsize,
        )
        buffer = bytearray(buffer_bytes)

        bin_hdr = np.frombuffer(buffer, dtype=self.spec.binary_header.dtype)

        transforms = TransformPipeline()

        if self.spec.endianness == Endianness.BIG:
            little_endian = TransformFactory.create("byte_swap", Endianness.LITTLE)
            transforms.add_transform(little_endian)

        interpret_revision = TransformFactory.create("segy_revision")
        transforms.add_transform(interpret_revision)

        return HeaderArray(transforms.apply(bin_hdr))

    def _initialize_spec(self, spec: SegySpec | None) -> SegySpec:
        """Initialize the spec based on the settings and/or file contents."""
        if spec is None:
            logger.info("No spec provided, inferring standard from binary header.")
            scan_result = self._infer_spec()
            inferred_spec = get_segy_standard(scan_result.revision)
            inferred_spec.endianness = scan_result.endianness
            spec = inferred_spec
        if spec.endianness is None:
            logger.info("No endianness provided, inferring from binary header.")
            spec.endianness = self._infer_spec().endianness
        return spec

    def _infer_spec(self) -> SegyInferResult:
        """Scan the file and infer the endianness and SEG-Y standard."""
        logger.info("Scanning binary header to infer SEG-Y standard and endianness.")
        bin_header_buffer = self.fs.read_block(fn=self.url, offset=3200, length=400)
        endianness_action = infer_endianness(bin_header_buffer, self.settings)
        revision = interpret_revision(
            bin_header_buffer, endianness_action, self.settings
        )

        if endianness_action == EndiannessAction.REVERSE:
            logger.debug("File not machine endianness, will reverse endian.")
            byte_order = "big" if sys.byteorder == "little" else "little"
        else:
            logger.debug("File is same as machine endianness, will read as-is.")
            byte_order = sys.byteorder

        return SegyInferResult(
            endianness=Endianness(byte_order),
            revision=revision,
        )

    def _update_spec(self) -> None:
        """Parse the binary header and apply some rules."""
        ext_text_header_spec = self.spec.ext_text_header

        if ext_text_header_spec is not None:
            num_ext_text = self.binary_header["num_extended_text_headers"].item()
            logger.info("Ext text headers, count from binary header: %s.", num_ext_text)
            ext_text_header_spec.count = num_ext_text

            settings_num_ext_text = self.settings.binary.ext_text_header
            if settings_num_ext_text is not None:
                logger.info(
                    "Using provided ext text header count from settings: %s",
                    settings_num_ext_text,
                )
                ext_text_header_spec.count = settings_num_ext_text

        sample_format_value = self.binary_header["data_sample_format"].item()
        data_sample_format_code = DataSampleFormatCode(sample_format_value)
        sample_format = ScalarType[data_sample_format_code.name]

        trace_data_spec = self.spec.trace.data
        trace_data_spec.format = sample_format
        trace_data_spec.samples = self.binary_header["samples_per_trace"].item()
        trace_data_spec.interval = self.binary_header["sample_interval"].item()

        logger.debug("Parsed sample format: %s", trace_data_spec.format)
        logger.debug("Parsed samples per trace: %s", trace_data_spec.samples)
        logger.debug("Parsed sample interval: %s", trace_data_spec.interval)

        self.spec.update_offsets()

        trace_offset = cast("int", self.spec.trace.offset)  # we know for sure not None
        trace_itemsize = self.spec.trace.itemsize
        trace_count = (self.file_size - trace_offset) // trace_itemsize

        # Ensure trace count and all other offsets match file size
        logger.debug("Calculated trace count: %s", trace_count)
        inferred_size = trace_offset + trace_count * trace_itemsize
        actual_size = self.file_size
        if inferred_size != actual_size:
            msg = f"{actual_size=} doesn't match parsed spec: {inferred_size} ."
            logger.error(msg)
            raise SegyFileSpecMismatchError(msg)

        self.spec.trace.count = trace_count

    @property
    def sample(self) -> AbstractIndexer:
        """Way to access the file to fetch trace data only."""
        return DataIndexer(
            self.fs,
            self.url,
            self.spec.trace,
            self.num_traces,
            transform_pipeline=self.accessors.sample_decode_pipeline,
        )

    @property
    def header(self) -> HeaderIndexer:
        """Way to access the file to fetch trace headers only."""
        return HeaderIndexer(
            self.fs,
            self.url,
            self.spec.trace,
            self.num_traces,
            transform_pipeline=self.accessors.header_decode_pipeline,
        )

    @property
    def trace(self) -> TraceIndexer:
        """Way to access the file to fetch full traces (headers + data)."""
        return TraceIndexer(
            self.fs,
            self.url,
            self.spec.trace,
            self.num_traces,
            transform_pipeline=self.accessors.trace_decode_pipeline,
        )



"""Data model implementations for trace specification."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

import numpy as np
from pydantic import Field
from pydantic import model_validator

from segy.schema.base import BaseDataType

if TYPE_CHECKING:
    from segy.schema.base import Endianness
    from segy.schema.format import ScalarType
    from segy.schema.header import HeaderSpec


class TraceDataSpec(BaseDataType):
    """A spec class for trace data (samples)."""

    format: ScalarType = Field(..., description="Format of trace samples.")  # noqa: A003
    samples: int | None = Field(
        default=None,
        description=(
            "Number of samples in trace. It can be variable, "
            "then it must be read from each trace header."
        ),
    )
    interval: int | None = Field(default=None, description="Sample interval of traces")

    @property
    def dtype(self) -> np.dtype[Any]:
        """Get numpy dtype with # of samples."""
        dtype = (self.format.dtype, (self.samples,))
        return np.dtype(dtype)


class TraceSpec(BaseDataType):
    """A spec class for a trace (header + data)."""

    header: HeaderSpec = Field(..., description="Trace header spec.")
    ext_header: HeaderSpec | None = Field(
        default=None, description="Extended trace header spec."
    )
    data: TraceDataSpec = Field(..., description="Trace data spec.")
    offset: int | None = Field(
        default=None, description="Starting offset of the trace."
    )
    endianness: Endianness | None = Field(
        default=None, description="Endianness of traces and headers."
    )
    count: int | None = Field(default=None, ge=0, description="Number of traces.")

    @model_validator(mode="after")
    def update_submodel_endianness(self) -> TraceSpec:
        """Ensure that submodel endianness matches the trace endianness."""
        self.header.endianness = self.endianness

        if self.ext_header is not None:
            self.ext_header.endianness = self.endianness

        return self

    @property
    def dtype(self) -> np.dtype[Any]:
        """Get numpy dtype."""
        header_dtype = self.header.dtype
        data_dtype = self.data.dtype

        trace_dtype = np.dtype([("header", header_dtype), ("data", data_dtype)])

        if self.endianness is not None:
            trace_dtype = trace_dtype.newbyteorder(self.endianness.symbol)

        return trace_dtype



"""Metadata and attribute encoding for headers in the SEG-Y standard."""

from enum import IntEnum


class SegyEndianCode(IntEnum):
    """Enumeration for endianness indicator from raw SEG-Y bytes.

    SEG-Y Revision 2 defines byte range 3297-3300 with an integer
    constant (32-bit) for unambiguous detection of file endianness.
    There are three cases. If it reads the value:
    - 0: Unknown pre-rev2, assume big-endian.
    - 1690906010: Don't swap bytes
    - 6730598510: Reverse bytes
    - 3362099510: Swap consecutive bytes
    """

    LEGACY = 0x00_00_00_00
    NATIVE = 0x01_02_03_04
    REVERSE = 0x04_03_02_01
    PAIRWISE_SWAP = 0x02_01_04_03


class DataSampleFormatCode(IntEnum):
    """Trace data type (format) codes for SEG-Y."""

    IBM32 = 1
    INT32 = 2
    INT16 = 3
    FLOAT32 = 5
    FLOAT64 = 6
    INT8 = 8
    INT64 = 9
    UINT32 = 10
    UINT16 = 11
    UINT64 = 12
    UINT8 = 16


class TraceSortingCode(IntEnum):
    """Trace sorting codes for SEG-Y trace headers."""

    OTHER = -1
    UNKNOWN = 0
    AS_RECORDED = 1
    CDP_ENSEMBLE = 2
    SINGLE_FOLD = 3
    HORIZONTALLY_STACKED = 4
    COMMON_SOURCE_POINT = 5
    COMMON_RECEIVER_POINT = 6
    COMMON_OFFSET_POINT = 7
    COMMOND_MID_POINT = 8
    COMMON_CONVERSION_POINT = 9


class SweepTypeCode(IntEnum):
    """Enumeration for types of sweep codes."""

    LINEAR = 1
    PARABOLIC = 2
    EXPONENTIAL = 3
    OTHER = 4


class SweepTaperType(IntEnum):
    """Enumeration for types of sweep taper codes."""

    LINEAR = 1
    COSINE_SQUARED = 2
    OTHER = 3


class CorrelatedTraces(IntEnum):
    """Enumeration for correlated traces."""

    NO = 1
    YES = 2


class BinaryGainRecovered(IntEnum):
    """Enumeration for binary gain recovery."""

    YES = 1
    NO = 2


class AmplitudeRecoverMethod(IntEnum):
    """Enumeration for amplitude recovery method."""

    NONE = 1
    SPHERICAL_DIVERGENCE = 2
    AGC = 3
    OTHER = 4


class MeasurementSystem(IntEnum):
    """Enumeration for units used in measurements."""

    METERS = 1
    FEET = 2



"""'The' SEG-Y implementation."""

from importlib import metadata

from segy.factory import SegyFactory
from segy.file import SegyFile

__all__ = ["SegyFile", "SegyFactory"]

try:
    __version__ = metadata.version("segy")
except metadata.PackageNotFoundError:
    __version__ = "unknown"



"""Main entrypoint to the segy cli."""

import typer

from segy.cli import dump as dump_command

app = typer.Typer(
    name="segy", no_args_is_help=True, pretty_exceptions_show_locals=False
)

app.add_typer(dump_command.app)



"""SEG-Y library exceptions."""


class SegyError(Exception):
    """Base class for all exceptions in this library."""


class SegyFileSpecMismatchError(SegyError):
    """Raised when file spec with parsed fields don't match file size."""


class EndiannessInferenceError(SegyError):
    """Raised when endianness inference has failed."""


class InvalidFieldError(SegyError):
    """Raised when a header key does not match known values."""

    def __init__(self, key: str, suggestions: list[str]):
        self.key = key
        self.suggestions = suggestions
        msg = self._generate_message()
        super().__init__(msg)

    def _generate_message(self) -> str:
        """Generate custom user message."""
        first, second, third = self.suggestions
        suggestion_text = f"'{first}', '{second}', or '{third}'"
        return f"Invalid key '{self.key}'. Did you mean one of: {suggestion_text}?"


class NonSpecFieldError(SegyError):
    """Raised when a header field is missing from a HeaderArray."""



"""Core functionality for SEG-Y ninja templates."""

from __future__ import annotations

from abc import abstractmethod
from enum import StrEnum
from typing import TYPE_CHECKING
from typing import Literal
from typing import cast

from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic.alias_generators import to_camel

if TYPE_CHECKING:
    from typing import Any

    import numpy as np


class CamelCaseModel(BaseModel):
    """Base model with camel case aliases. Extends BaseModel.

    Attributes:
        model_config: The configuration dictionary for the model.
            - alias_generator: The function used to generate aliases.
            - populate_by_name: Flag indicating whether to populate by name.
    """

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        validate_assignment=True,
    )

    def model_dump(self, *args: Any, **kwargs: Any) -> dict[str, Any]:  # noqa: ANN401
        """Dump the model into a dictionary by alias."""
        return super().model_dump(*args, **kwargs, by_alias=True)

    def model_dump_json(self, *args: Any, **kwargs: Any) -> str:  # noqa: ANN401
        """Dump the model into a JSON string by alias."""
        return super().model_dump_json(*args, **kwargs, by_alias=True)

    def _repr_json_(self) -> dict[str, Any]:
        """Return JSON-able dictionary form of model to render in notebooks."""
        return self.model_dump(mode="json")


class BaseDataType(CamelCaseModel):
    """A base model for all SEG-Y Ninja types."""

    @property
    @abstractmethod
    def dtype(self) -> np.dtype[Any]:
        """Abstract property for subclasses to return the numpy dtype."""

    @property
    def itemsize(self) -> int:
        """Number of bytes for the data type."""
        return self.dtype.itemsize


class Endianness(StrEnum):
    """Enumeration class with three possible endianness values.

    Attributes:
        BIG: Big endian.
        LITTLE: Little endian.

    Examples:
        >>> endian = Endianness.BIG
        >>> print(endian.symbol)
        >
    """

    BIG = "big"
    LITTLE = "little"

    def __init__(self, _: str) -> None:
        self._symbol_map = {
            "big": ">",
            "little": "<",
        }

    @property
    def symbol(self) -> Literal["<", ">", "="]:
        """Get the numpy symbol for the endianness from mapping."""
        return cast("Literal['<', '>', '=']", self._symbol_map[self.value])



"""SEG-Y binary header definitions."""

from segy.standards.fields.base import SegStandardEnum


# fmt: off
class Rev0(SegStandardEnum):
    """Definition of SEG-Y Rev0 binary headers."""

    JOB_ID                            = (1, "int32")
    LINE_NUM                          = (5, "int32")
    REEL_NUM                          = (9, "int32")
    DATA_TRACES_PER_ENSEMBLE          = (13, "int16")
    AUX_TRACES_PER_ENSEMBLE           = (15, "int16")
    SAMPLE_INTERVAL                   = (17, "int16")
    ORIG_SAMPLE_INTERVAL              = (19, "int16")
    SAMPLES_PER_TRACE                 = (21, "int16")
    ORIG_SAMPLES_PER_TRACE            = (23, "int16")
    DATA_SAMPLE_FORMAT                = (25, "int16")
    ENSEMBLE_FOLD                     = (27, "int16")
    TRACE_SORTING_CODE                = (29, "int16")
    VERTICAL_SUM_CODE                 = (31, "int16")
    SWEEP_FREQ_START                  = (33, "int16")
    SWEEP_FREQ_END                    = (35, "int16")
    SWEEP_LENGTH                      = (37, "int16")
    SWEEP_TYPE_CODE                   = (39, "int16")
    SWEEP_TRACE_NUM                   = (41, "int16")
    SWEEP_TAPER_START                 = (43, "int16")
    SWEEP_TAPER_END                   = (45, "int16")
    TAPER_TYPE_CODE                   = (47, "int16")
    CORRELATED_DATA_CODE              = (49, "int16")
    BINARY_GAIN_CODE                  = (51, "int16")
    AMP_RECOVERY_CODE                 = (53, "int16")
    MEASUREMENT_SYSTEM_CODE           = (55, "int16")
    IMPULSE_POLARITY_CODE             = (57, "int16")
    VIBRATORY_POLARITY_CODE           = (59, "int16")


class Rev1(SegStandardEnum):
    """Definition of SEG-Y Rev1 binary headers."""

    SEGY_REVISION                     = (301, "int16")
    FIXED_LENGTH_TRACE_FLAG           = (303, "int16")
    NUM_EXTENDED_TEXT_HEADERS         = (305, "int16")


class Rev2(SegStandardEnum):
    """Definition of SEG-Y Rev2 binary headers."""

    EXTENDED_DATA_TRACES_PER_ENSEMBLE = (61, "int32")
    EXTENDED_AUX_TRACES_PER_ENSEMBLE  = (65, "int32")
    EXTENDED_SAMPLES_PER_TRACE        = (69, "int32")
    EXTENDED_SAMPLE_INTERVAL          = (73, "float64")
    EXTENDED_ORIG_SAMPLE_INTERVAL     = (81, "float64")
    EXTENDED_ORIG_SAMPLES_PER_TRACE   = (89, "int32")
    EXTENDED_ENSEMBLE_FOLD            = (93, "int32")
    BYTE_ORDER                        = (97, "int32")
    SEGY_REVISION_MAJOR               = (301, "uint8")
    SEGY_REVISION_MINOR               = (302, "uint8")
    MAX_EXTENDED_TRACE_HEADERS        = (307, "int32")
    TIME_BASIS_CODE                   = (311, "int16")
    NUM_TRACES                        = (313, "uint64")
    BYTE_OFFSET_FIRST_TRACE           = (321, "uint64")
    NUM_DATA_TRAILER_STANZAS          = (329, "int32")


class Rev21(SegStandardEnum):
    """Definition of SEG-Y Rev2 binary headers."""

    MAX_EXTENDED_TRACE_HEADERS        = (307, "int16")
    SURVEY_TYPE                       = (309, "int16")
# fmt:on



"""Base tuple and enum definitions for SEG-Y standard definitions."""

from __future__ import annotations

from collections import namedtuple
from enum import Enum

from segy.schema import HeaderField

FieldTuple = namedtuple("FieldTuple", ["byte", "format"])


class SegStandardEnum(FieldTuple, Enum):
    """A special enum for defining standards with data model generation attribute."""

    def __repr__(self) -> str:
        """Custom concise repr."""
        cls_name = self.__class__.__name__
        val_name = self._name_
        val_repr = f"byte={self.byte}, format={self.format}"
        return f"<{cls_name}.{val_name}: {val_repr}>"

    def __new__(cls, byte: int, format: str) -> SegStandardEnum:  # noqa: A002
        """This will ensure any tuple is converted to FieldTuple."""
        return FieldTuple.__new__(cls, byte, format)

    @property
    def model(self) -> HeaderField:
        """Generate proper data model for header field with validation."""
        return HeaderField(name=self.name.lower(), byte=self.byte, format=self.format)



"""SEG-Y Revision 0 Specification."""

from segy.schema.base import Endianness
from segy.schema.format import ScalarType
from segy.schema.format import TextHeaderEncoding
from segy.schema.header import HeaderSpec
from segy.schema.segy import SegySpec
from segy.schema.segy import SegyStandard
from segy.schema.text_header import ExtendedTextHeaderSpec
from segy.schema.text_header import TextHeaderSpec
from segy.schema.trace import TraceDataSpec
from segy.schema.trace import TraceSpec
from segy.standards.fields import binary
from segy.standards.fields import trace

BIN_HDR_FIELDS_REV0 = [field.model for field in binary.Rev0]
TRC_HDR_FIELDS_REV0 = [field.model for field in trace.Rev0]

BIN_HDR_FIELDS_REV1 = BIN_HDR_FIELDS_REV0 + [field.model for field in binary.Rev1]
TRC_HDR_FIELDS_REV1 = TRC_HDR_FIELDS_REV0 + [field.model for field in trace.Rev1]

# Rev2 removes segy_revision and splits it into major/minor revision fields
BIN_HDR_FIELDS_REV2 = BIN_HDR_FIELDS_REV1 + [field.model for field in binary.Rev2]
BIN_HDR_FIELDS_REV2.remove(binary.Rev1.SEGY_REVISION.model)
BIN_HDR_FIELDS_REV2 = sorted(BIN_HDR_FIELDS_REV2, key=lambda f: f.byte)
TRC_HDR_FIELDS_REV2 = TRC_HDR_FIELDS_REV1 + [field.model for field in trace.Rev2]

# Rev2.1 splits maximum extended trace headers into two 16-bit.
# First reduces it from 32-bit to 16-bit
# Second gets new "survey type" field
BIN_HDR_FIELDS_REV21 = BIN_HDR_FIELDS_REV2 + [field.model for field in binary.Rev21]
BIN_HDR_FIELDS_REV21.remove(binary.Rev2.MAX_EXTENDED_TRACE_HEADERS.model)
BIN_HDR_FIELDS_REV21 = sorted(BIN_HDR_FIELDS_REV21, key=lambda f: f.byte)
TRC_HDR_FIELDS_REV21 = TRC_HDR_FIELDS_REV2  # no change


text_header_ebcdic = TextHeaderSpec(
    rows=40,
    cols=80,
    encoding=TextHeaderEncoding.EBCDIC,
)

ext_text_header_ebcdic_3200 = ExtendedTextHeaderSpec(spec=text_header_ebcdic)

REV0 = SegySpec(
    segy_standard=SegyStandard.REV0,
    endianness=Endianness.BIG,
    text_header=text_header_ebcdic,
    binary_header=HeaderSpec(fields=BIN_HDR_FIELDS_REV0, item_size=400, offset=3200),
    trace=TraceSpec(
        data=TraceDataSpec(format=ScalarType.IBM32),
        header=HeaderSpec(fields=TRC_HDR_FIELDS_REV0, item_size=240),
    ),
)

REV1 = SegySpec(
    segy_standard=SegyStandard.REV1,
    endianness=Endianness.BIG,
    text_header=text_header_ebcdic,
    binary_header=HeaderSpec(fields=BIN_HDR_FIELDS_REV1, item_size=400, offset=3200),
    ext_text_header=ext_text_header_ebcdic_3200,
    trace=TraceSpec(
        data=TraceDataSpec(format=ScalarType.IBM32),
        header=HeaderSpec(fields=TRC_HDR_FIELDS_REV1, item_size=240),
    ),
)

REV2 = SegySpec(
    segy_standard=SegyStandard.REV2,
    endianness=Endianness.BIG,
    text_header=text_header_ebcdic,
    binary_header=HeaderSpec(fields=BIN_HDR_FIELDS_REV2, item_size=400, offset=3200),
    ext_text_header=ext_text_header_ebcdic_3200,
    trace=TraceSpec(
        data=TraceDataSpec(format=ScalarType.IBM32),
        header=HeaderSpec(fields=TRC_HDR_FIELDS_REV2, item_size=240),
    ),
)

REV21 = SegySpec(
    segy_standard=SegyStandard.REV21,
    endianness=Endianness.BIG,
    text_header=text_header_ebcdic,
    binary_header=HeaderSpec(fields=BIN_HDR_FIELDS_REV21, item_size=400, offset=3200),
    ext_text_header=ext_text_header_ebcdic_3200,
    trace=TraceSpec(
        data=TraceDataSpec(format=ScalarType.IBM32),
        header=HeaderSpec(fields=TRC_HDR_FIELDS_REV21, item_size=240),
    ),
)



"""Segyio binary and trace header mappings.

We added segyio header name mappings to canonical header names we define
for SEG-Y headers.

segyio implements Rev0, Rev1, and some Rev2 header fields in their standard.
We have mapped everything to the forms we defined.

Sources:
https://segyio.readthedocs.io/en/latest/segyio.html#trace-header-keys
https://segyio.readthedocs.io/en/latest/segyio.html#binary-header-keys
"""

from segy.standards.fields import binary
from segy.standards.fields import trace

# fmt: off
SEGYIO_BIN_FIELD_MAP = {
    "JobID":                 binary.Rev0.JOB_ID,
    "LineNumber":            binary.Rev0.LINE_NUM,
    "ReelNumber":            binary.Rev0.REEL_NUM,
    "Traces":                binary.Rev0.DATA_TRACES_PER_ENSEMBLE,
    "AuxTraces":             binary.Rev0.AUX_TRACES_PER_ENSEMBLE,
    "Interval":              binary.Rev0.SAMPLE_INTERVAL,
    "IntervalOriginal":      binary.Rev0.ORIG_SAMPLE_INTERVAL,
    "Samples":               binary.Rev0.SAMPLES_PER_TRACE,
    "SamplesOriginal":       binary.Rev0.ORIG_SAMPLES_PER_TRACE,
    "Format":                binary.Rev0.DATA_SAMPLE_FORMAT,
    "EnsembleFold":          binary.Rev0.ENSEMBLE_FOLD,
    "SortingCode":           binary.Rev0.TRACE_SORTING_CODE,
    "VerticalSum":           binary.Rev0.VERTICAL_SUM_CODE,
    "SweepFrequencyStart":   binary.Rev0.SWEEP_FREQ_START,
    "SweepFrequencyEnd":     binary.Rev0.SWEEP_FREQ_END,
    "SweepLength":           binary.Rev0.SWEEP_LENGTH,
    "Sweep":                 binary.Rev0.SWEEP_TYPE_CODE,
    "SweepChannel":          binary.Rev0.SWEEP_TRACE_NUM,
    "SweepTaperStart":       binary.Rev0.SWEEP_TAPER_START,
    "SweepTaperEnd":         binary.Rev0.SWEEP_TAPER_END,
    "Taper":                 binary.Rev0.TAPER_TYPE_CODE,
    "CorrelatedTraces":      binary.Rev0.CORRELATED_DATA_CODE,
    "BinaryGainRecovery":    binary.Rev0.BINARY_GAIN_CODE,
    "AmplitudeRecovery":     binary.Rev0.AMP_RECOVERY_CODE,
    "MeasurementSystem":     binary.Rev0.MEASUREMENT_SYSTEM_CODE,
    "ImpulseSignalPolarity": binary.Rev0.IMPULSE_POLARITY_CODE,
    "VibratoryPolarity":     binary.Rev0.VIBRATORY_POLARITY_CODE,
    "ExtTraces":             binary.Rev2.EXTENDED_DATA_TRACES_PER_ENSEMBLE,
    "ExtAuxTraces":          binary.Rev2.EXTENDED_AUX_TRACES_PER_ENSEMBLE,
    "ExtSamples":            binary.Rev2.EXTENDED_SAMPLES_PER_TRACE,
    "ExtSamplesOriginal":    binary.Rev2.EXTENDED_ORIG_SAMPLES_PER_TRACE,
    "ExtEnsembleFold":       binary.Rev2.EXTENDED_ENSEMBLE_FOLD,
    "SEGYRevision":          binary.Rev2.SEGY_REVISION_MAJOR,
    "SEGYRevisionMinor":     binary.Rev2.SEGY_REVISION_MINOR,
    "TraceFlag":             binary.Rev1.FIXED_LENGTH_TRACE_FLAG,
    "ExtendedHeaders":       binary.Rev1.NUM_EXTENDED_TEXT_HEADERS,
}

SEGYIO_TRACE_FIELD_MAP = {
    "TRACE_SEQUENCE_LINE":                    trace.Rev0.TRACE_SEQ_NUM_LINE,
    "TRACE_SEQUENCE_FILE":                    trace.Rev0.TRACE_SEQ_NUM_REEL,
    "FieldRecord":                            trace.Rev0.ORIG_FIELD_RECORD_NUM,
    "TraceNumber":                            trace.Rev0.TRACE_NUM_ORIG_RECORD,
    "EnergySourcePoint":                      trace.Rev0.ENERGY_SOURCE_POINT_NUM,
    "CDP":                                    trace.Rev0.ENSEMBLE_NUM,
    "CDP_TRACE":                              trace.Rev0.TRACE_NUM_ENSEMBLE,
    "TraceIdentificationCode":                trace.Rev0.TRACE_ID_CODE,
    "NSummedTraces":                          trace.Rev0.VERTICALLY_SUMMED_TRACES,
    "NStackedTraces":                         trace.Rev0.HORIZONTALLY_STACKED_TRACES,
    "DataUse":                                trace.Rev0.DATA_USE,
    "offset":                                 trace.Rev0.SOURCE_TO_RECEIVER_DISTANCE,
    "ReceiverGroupElevation":                 trace.Rev0.RECEIVER_GROUP_ELEVATION,
    "SourceSurfaceElevation":                 trace.Rev0.SOURCE_SURFACE_ELEVATION,
    "SourceDepth":                            trace.Rev0.SOURCE_DEPTH_BELOW_SURFACE,
    "ReceiverDatumElevation":                 trace.Rev0.RECEIVER_DATUM_ELEVATION,
    "SourceDatumElevation":                   trace.Rev0.SOURCE_DATUM_ELEVATION,
    "SourceWaterDepth":                       trace.Rev0.SOURCE_WATER_DEPTH,
    "GroupWaterDepth":                        trace.Rev0.RECEIVER_WATER_DEPTH,
    "ElevationScalar":                        trace.Rev0.ELEVATION_DEPTH_SCALAR,
    "SourceGroupScalar":                      trace.Rev0.COORDINATE_SCALAR,
    "SourceX":                                trace.Rev0.SOURCE_COORD_X,
    "SourceY":                                trace.Rev0.SOURCE_COORD_Y,
    "GroupX":                                 trace.Rev0.GROUP_COORD_X,
    "GroupY":                                 trace.Rev0.GROUP_COORD_Y,
    "CoordinateUnits":                        trace.Rev0.COORDINATE_UNIT,
    "WeatheringVelocity":                     trace.Rev0.WEATHERING_VELOCITY,
    "SubWeatheringVelocity":                  trace.Rev0.SUBWEATHERING_VELOCITY,
    "SourceUpholeTime":                       trace.Rev0.SOURCE_UPHOLE_TIME,
    "GroupUpholeTime":                        trace.Rev0.GROUP_UPHOLE_TIME,
    "SourceStaticCorrection":                 trace.Rev0.SOURCE_STATIC_CORRECTION,
    "GroupStaticCorrection":                  trace.Rev0.RECEIVER_STATIC_CORRECTION,
    "TotalStaticApplied":                     trace.Rev0.TOTAL_STATIC_APPLIED,
    "LagTimeA":                               trace.Rev0.LAG_TIME_A,
    "LagTimeB":                               trace.Rev0.LAG_TIME_B,
    "DelayRecordingTime":                     trace.Rev0.DELAY_RECORDING_TIME,
    "MuteTimeStart":                          trace.Rev0.MUTE_TIME_START,
    "MuteTimeEND":                            trace.Rev0.MUTE_TIME_END,
    "TRACE_SAMPLE_COUNT":                     trace.Rev0.SAMPLES_PER_TRACE,
    "TRACE_SAMPLE_INTERVAL":                  trace.Rev0.SAMPLE_INTERVAL,
    "GainType":                               trace.Rev0.INSTRUMENT_GAIN_TYPE,
    "InstrumentGainConstant":                 trace.Rev0.INSTRUMENT_GAIN_CONST,
    "InstrumentInitialGain":                  trace.Rev0.INSTRUMENT_GAIN_INITIAL,
    "Correlated":                             trace.Rev0.CORRELATED_DATA,
    "SweepFrequencyStart":                    trace.Rev0.SWEEP_FREQ_START,
    "SweepFrequencyEnd":                      trace.Rev0.SWEEP_FREQ_END,
    "SweepLength":                            trace.Rev0.SWEEP_LENGTH,
    "SweepType":                              trace.Rev0.SWEEP_TYPE,
    "SweepTraceTaperLengthStart":             trace.Rev0.SWEEP_TAPER_START,
    "SweepTraceTaperLengthEnd":               trace.Rev0.SWEEP_TAPER_END,
    "TaperType":                              trace.Rev0.TAPER_TYPE,
    "AliasFilterFrequency":                   trace.Rev0.ALIAS_FILTER_FREQ,
    "AliasFilterSlope":                       trace.Rev0.ALIAS_FILTER_SLOPE,
    "NotchFilterFrequency":                   trace.Rev0.NOTCH_FILTER_FREQ,
    "NotchFilterSlope":                       trace.Rev0.NOTCH_FILTER_SLOPE,
    "LowCutFrequency":                        trace.Rev0.LOW_CUT_FREQ,
    "HighCutFrequency":                       trace.Rev0.HIGH_CUT_FREQ,
    "LowCutSlope":                            trace.Rev0.LOW_CUT_SLOPE,
    "HighCutSlope":                           trace.Rev0.HIGH_CUT_SLOPE,
    "YearDataRecorded":                       trace.Rev0.YEAR_RECORDED,
    "DayOfYear":                              trace.Rev0.DAY_OF_YEAR,
    "HourOfDay":                              trace.Rev0.HOUR_OF_DAY,
    "MinuteOfHour":                           trace.Rev0.MINUTE_OF_HOUR,
    "SecondOfMinute":                         trace.Rev0.SECOND_OF_MINUTE,
    "TimeBaseCode":                           trace.Rev0.TIME_BASIS_CODE,
    "TraceWeightingFactor":                   trace.Rev0.TRACE_WEIGHTING_FACTOR,
    "GeophoneGroupNumberRoll1":               trace.Rev0.GROUP_NUM_ROLL_SWITCH,
    "GeophoneGroupNumberFirstTraceOrigField": trace.Rev0.GROUP_NUM_FIRST_TRACE,
    "GeophoneGroupNumberLastTraceOrigField":  trace.Rev0.GROUP_NUM_LAST_TRACE,
    "GapSize":                                trace.Rev0.GAP_SIZE,
    "OverTravel":                             trace.Rev0.TAPER_OVERTRAVEL,
    "CDP_X":                                  trace.Rev1.CDP_X,
    "CDP_Y":                                  trace.Rev1.CDP_Y,
    "INLINE_3D":                              trace.Rev1.INLINE,
    "CROSSLINE_3D":                           trace.Rev1.CROSSLINE,
    "ShotPoint":                              trace.Rev1.SHOT_POINT,
    "ShotPointScalar":                        trace.Rev1.SHOT_POINT_SCALAR,
    "TraceValueMeasurementUnit":              trace.Rev1.TRACE_VALUE_UNIT,
    "TransductionConstantMantissa":           trace.Rev1.TRANSDUCTION_CONST_MANTISSA,
    "TransductionConstantPower":              trace.Rev1.TRANSDUCTION_CONST_EXPONENT,
    "TransductionUnit":                       trace.Rev1.TRANSDUCTION_UNIT,
    "TraceIdentifier":                        trace.Rev1.DEVICE_TRACE_ID,
    "ScalarTraceHeader":                      trace.Rev1.TIMES_SCALAR,
    "SourceType":                             trace.Rev1.SOURCE_TYPE_ORIENTATION,
    "SourceEnergyDirectionMantissa":          trace.Rev1.SOURCE_ENERGY_DIR_MANTISSA,
    "SourceEnergyDirectionExponent":          trace.Rev1.SOURCE_ENERGY_DIR_EXPONENT,
    "SourceMeasurementMantissa":              trace.Rev1.SOURCE_MEASUREMENT_MANTISSA,
    "SourceMeasurementExponent":              trace.Rev1.SOURCE_MEASUREMENT_EXPONENT,
    "SourceMeasurementUnit":                  trace.Rev1.SOURCE_MEASUREMENT_UNIT,
}
# fmt: on



"""SEG-Y file specification with minimal, mandatory fields."""

from segy.schema.base import Endianness
from segy.schema.format import ScalarType
from segy.schema.format import TextHeaderEncoding
from segy.schema.header import HeaderSpec
from segy.schema.segy import SegySpec
from segy.schema.text_header import TextHeaderSpec
from segy.schema.trace import TraceDataSpec
from segy.schema.trace import TraceSpec
from segy.standards.fields import binary

BIN_HDR_FIELDS = [
    binary.Rev0.SAMPLE_INTERVAL,
    binary.Rev0.ORIG_SAMPLE_INTERVAL,
    binary.Rev0.SAMPLES_PER_TRACE,
    binary.Rev0.ORIG_SAMPLES_PER_TRACE,
    binary.Rev0.DATA_SAMPLE_FORMAT,
    binary.Rev1.SEGY_REVISION,
    binary.Rev1.FIXED_LENGTH_TRACE_FLAG,
    binary.Rev1.NUM_EXTENDED_TEXT_HEADERS,
]
BIN_HDR_FIELD_MODELS = [field.model for field in BIN_HDR_FIELDS]

textual_file_header = TextHeaderSpec(
    rows=40,
    cols=80,
    offset=0,
    encoding=TextHeaderEncoding.EBCDIC,
)

minimal_segy = SegySpec(
    segy_standard=None,
    text_header=textual_file_header,
    binary_header=HeaderSpec(fields=BIN_HDR_FIELD_MODELS, item_size=400, offset=3200),
    trace=TraceSpec(
        header=HeaderSpec(fields=[], item_size=240),
        data=TraceDataSpec(format=ScalarType.IBM32),
    ),
    endianness=Endianness.BIG,
)
